from autosar.writer.writer_base import ElementWriter
import autosar.portinterface

class XMLPortInterfaceWriter(ElementWriter):
    def __init__(self,version, patch):
        super().__init__(version, patch)

        if self.version >= 3.0 and self.version < 4.0:
            self.switcher = {
                             'SoftwareAddressMethod': self.writeSoftwareAddressMethodXML,
                             'SenderReceiverInterface': self.writeSenderReceiverInterfaceXML,
                             'ParameterInterface': self.writeCalPrmInterfaceXML,
                             'ClientServerInterface': self.writeClientServerInterfaceXML,

            }
        elif self.version >= 4.0:
            self.switcher = {
                              'ClientServerInterface': self.writeClientServerInterfaceXML,
                              'ModeSwitchInterface': self.writeModeSwitchInterfaceXML,
                              'SenderReceiverInterface': self.writeSenderReceiverInterfaceXML,
                              'ParameterInterface': self.writeParameterInterfaceXML,
                              'SoftwareAddressMethod': self.writeSoftwareAddressMethodXML,
                              'NvDataInterface': self.writeNvDataInterfaceXML,
            }
        else:
            switch.keys = {}

    def getSupportedXML(self):
        return self.switcher.keys()

    def getSupportedCode(self):
        return []

    def writeElementXML(self, elem):
        xmlWriteFunc = self.switcher.get(type(elem).__name__)
        if xmlWriteFunc is not None:
            return xmlWriteFunc(elem)
        else:
            return None

    def writeElementCode(self, elem, localvars):
        raise NotImplementedError('writeElementCode')

    def writeSenderReceiverInterfaceXML(self, portInterface):
        assert(isinstance(portInterface,autosar.portinterface.SenderReceiverInterface))
        ws = portInterface.rootWS()
        lines=[]
        lines.append('<SENDER-RECEIVER-INTERFACE>')
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%portInterface.name,1))
        descLines = self.writeDescXML(portInterface)
        if descLines is not None:
            lines.extend(self.indent(descLines,1))
        if portInterface.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(portInterface.adminData),1))
        lines.append(self.indent('<IS-SERVICE>%s</IS-SERVICE>'%self.toBooleanStr(portInterface.isService),1))
        if (self.version >= 4.0) and (portInterface.serviceKind is not None):
            lines.append(self.indent('<SERVICE-KIND>%s</SERVICE-KIND>'%portInterface.serviceKind,1))
        if len(portInterface.dataElements)>0:
            lines.append(self.indent('<DATA-ELEMENTS>',1))
            for elem in portInterface.dataElements:
                lines.extend(self.indent(self.writeDataElementXML(elem),2))
            lines.append(self.indent('</DATA-ELEMENTS>',1))
        else:
            lines.append(self.indent('<DATA-ELEMENTS/>',1))
        if len(portInterface.modeGroups) > 0:
            lines.append(self.indent('<MODE-GROUPS>',1))
            for group in portInterface.modeGroups:
                lines.extend(self.indent(self.writeModeGroupXML(group),2))
            lines.append(self.indent('</MODE-GROUPS>',1))
        if len(portInterface.invalidationPolicies)>0:
            lines.append(self.indent('<INVALIDATION-POLICYS>',1))
            for invalidationPolicy in portInterface.invalidationPolicies:
                lines.extend(self.indent(self.writeInvalidationPolicyXML(ws, invalidationPolicy),2))
            lines.append(self.indent('</INVALIDATION-POLICYS>',1))
        lines.append('</SENDER-RECEIVER-INTERFACE>')
        return lines

    def writeCalPrmInterfaceXML(self, portInterface):
        assert(isinstance(portInterface,autosar.portinterface.ParameterInterface))
        lines=[]
        lines.append('<CALPRM-INTERFACE>')
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%portInterface.name,1))
        lines.append(self.indent('<IS-SERVICE>%s</IS-SERVICE>'%self.toBooleanStr(portInterface.isService),1))
        if len(portInterface.parameters)>0:
            lines.append(self.indent('<CALPRM-ELEMENTS>',1))
            for elem in portInterface.parameters:
                lines.extend(self.indent(self.writeCalParamElementXML(elem),2))
            lines.append(self.indent('</CALPRM-ELEMENTS>',1))
        else:
            lines.append(self.indent('<CALPRM-ELEMENTS/>',1))
        lines.append('</CALPRM-INTERFACE>')
        return lines

    def writeCalParamElementXML(self,elem):
        assert(isinstance(elem,autosar.element.ParameterDataPrototype))
        lines=[]
        lines.append('<CALPRM-ELEMENT-PROTOTYPE>')
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%elem.name,1))
        ws = elem.rootWS()

        if elem.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(elem.adminData),1))
        if elem.swAddressMethodRef is not None:
            lines.append(self.indent('<SW-DATA-DEF-PROPS>',1))
            swAddrMethod = ws.find(elem.swAddressMethodRef)
            if (swAddrMethod is None):
                raise ValueError("invalid reference: '%s'"%elem.swAddressMethodRef)
            else:
                lines.append(self.indent('<SW-ADDR-METHOD-REF DEST="%s">%s</SW-ADDR-METHOD-REF>'%(swAddrMethod.tag(self.version),swAddrMethod.ref),2))
            lines.append(self.indent('</SW-DATA-DEF-PROPS>',1))
        typeElem = ws.find(elem.typeRef, role="DataType")
        if (typeElem is None):
            raise ValueError("invalid type reference: '%s'"%elem.typeRef)
        else:
            lines.append(self.indent('<TYPE-TREF DEST="%s">%s</TYPE-TREF>'%(typeElem.tag(self.version),typeElem.ref),1))
        lines.append('</CALPRM-ELEMENT-PROTOTYPE>')
        return lines

    def writeParameterInterfaceXML(self, portInterface):
        assert(isinstance(portInterface,autosar.portinterface.ParameterInterface))
        lines=[]
        ws = portInterface.rootWS()
        assert(ws is not None)
        lines.append('<%s>'%portInterface.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%portInterface.name,1))
        if portInterface.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(portInterface.adminData),1))
        if len(portInterface.parameters)>0:
            lines.append(self.indent('<PARAMETERS>',1))
            for elem in portInterface.parameters:
                lines.extend(self.indent(self._writeParameterElement(elem, ws),2))
            lines.append(self.indent('</PARAMETERS>',1))
        else:
            lines.append(self.indent('<PARAMETERS/>',1))
        lines.append('</%s>'%portInterface.tag(self.version))
        return lines

    def _writeParameterElement(self, parameter, ws):
        assert(isinstance(parameter, autosar.element.ParameterDataPrototype))
        lines=[]
        lines.append('<%s>'%parameter.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%parameter.name,1))

        lines.append(self.indent('<SW-DATA-DEF-PROPS>',1))
        if parameter.swCalibrationAccess is None:
            access = 'NOT-ACCESSIBLE'
        else:
            access = parameter.swCalibrationAccess
        variants = [autosar.base.SwDataDefPropsConditional(swAddressMethodRef=parameter.swAddressMethodRef, swCalibrationAccess=access)]
        lines.extend(self.indent(self.writeSwDataDefPropsVariantsXML(ws, variants),2))
        lines.append(self.indent('</SW-DATA-DEF-PROPS>',1))
        typeElem = ws.find(parameter.typeRef, role="DataType")
        if (typeElem is None):
            raise ValueError("invalid type reference: '%s'"%parameter.typeRef)
        else:
            lines.append(self.indent('<TYPE-TREF DEST="%s">%s</TYPE-TREF>'%(typeElem.tag(self.version),typeElem.ref),1))
        lines.append('</%s>'%parameter.tag(self.version))
        return lines

    def writeClientServerInterfaceXML(self, portInterface):
        assert(isinstance(portInterface,autosar.portinterface.ClientServerInterface))
        lines=[]
        lines.append('<CLIENT-SERVER-INTERFACE>')
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%portInterface.name,1))
        descLines = self.writeDescXML(portInterface)
        if descLines is not None:
            lines.extend(self.indent(descLines,1))
        if portInterface.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(portInterface.adminData),1))
        lines.append(self.indent('<IS-SERVICE>%s</IS-SERVICE>'%self.toBooleanStr(portInterface.isService),1))
        if (portInterface.serviceKind is not None) and (self.version >= 4.0):
            lines.append(self.indent('<SERVICE-KIND>%s</SERVICE-KIND>'%portInterface.serviceKind,1))
        if len(portInterface.operations)>0:
            lines.append(self.indent('<OPERATIONS>',1))
            for operation in portInterface.operations:
                lines.extend(self.indent(self.writeOperationXML(operation),2))
            lines.append(self.indent('</OPERATIONS>',1))
        else:
            lines.append(self.indent('<OPERATIONS/>',1))
        if len(portInterface.applicationErrors)>0:
            lines.append(self.indent('<POSSIBLE-ERRORS>',1))
            for applicationError in portInterface.applicationErrors:
                lines.extend(self.indent(self.writeApplicationErrorXML(applicationError),2))
            lines.append(self.indent('</POSSIBLE-ERRORS>',1))
        lines.append('</CLIENT-SERVER-INTERFACE>')

        return lines

    def writeOperationXML(self,operation):
        assert(isinstance(operation,autosar.portinterface.Operation))
        ws = operation.rootWS()
        assert(ws is not None)
        lines=[]
        lines.append('<%s>'%operation.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%operation.name,1))
        descLines = self.writeDescXML(operation)
        if descLines is not None:
            lines.extend(self.indent(descLines,1))
        if len(operation.arguments)>0:
            lines.append(self.indent('<ARGUMENTS>',1))
            for argument in operation.arguments:
                lines.extend(self.indent(self.writeArgumentXML(ws, argument),2))
            lines.append(self.indent('</ARGUMENTS>',1))
        if len(operation.errorRefs)>0:
            lines.append(self.indent('<POSSIBLE-ERROR-REFS>',1))
            for errorRef in operation.errorRefs:
                errorElem = ws.find(errorRef)
                if (errorElem is None):
                    raise ValueError("invalid error reference: '%s'"%errorRef)
                else:
                    lines.append(self.indent('<POSSIBLE-ERROR-REF DEST="%s">%s</POSSIBLE-ERROR-REF>'%(errorElem.tag(self.version),errorElem.ref),2))
            lines.append(self.indent('</POSSIBLE-ERROR-REFS>',1))
        lines.append('</%s>'%operation.tag(self.version))
        return lines

    def writeArgumentXML(self, ws, argument):
        assert(isinstance(argument, autosar.portinterface.Argument))
        lines=[]
        lines.append('<%s>'%argument.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%argument.name,1))
        if self.version >= 4.0:
            descLines = self.writeDescXML(argument)
            if descLines is not None:
                lines.extend(self.indent(descLines,1))
            lines.append(self.indent('<SW-DATA-DEF-PROPS>',1))
            if argument.swCalibrationAccess is None:
                tmp = 'NOT-ACCESSIBLE'
            else:
                tmp = argument.swCalibrationAccess
            variants = [autosar.base.SwDataDefPropsConditional(swCalibrationAccess=tmp)]
            lines.extend(self.indent(self.writeSwDataDefPropsVariantsXML(ws, variants),2))
            lines.append(self.indent('</SW-DATA-DEF-PROPS>',1))
        typeElem = ws.find(argument.typeRef, role="DataType")
        if (typeElem is None):
            raise ValueError("invalid type reference: '%s'"%argument.typeRef)
        else:
            lines.append(self.indent('<TYPE-TREF DEST="%s">%s</TYPE-TREF>'%(typeElem.tag(self.version),typeElem.ref),1))
        lines.append(self.indent('<DIRECTION>%s</DIRECTION>'%argument.direction,1))
        if self.version >= 4.0:
            if argument.serverArgumentImplPolicy is not None:
                policy = argument.serverArgumentImplPolicy
            else:
                policy = 'USE-ARGUMENT-TYPE'
            lines.append(self.indent('<SERVER-ARGUMENT-IMPL-POLICY>%s</SERVER-ARGUMENT-IMPL-POLICY>' % (policy), 1))
        lines.append('</%s>'%argument.tag(self.version))
        return lines


    def writeApplicationErrorXML(self,applicationError):
        assert(isinstance(applicationError, autosar.portinterface.ApplicationError))
        lines=[]
        lines.append('<%s>'%applicationError.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%applicationError.name,1))
        lines.append(self.indent('<ERROR-CODE>%d</ERROR-CODE>'%applicationError.errorCode,1))
        lines.append('</%s>'%applicationError.tag(self.version))
        return lines

    def writeModeGroupXML(self,modeGroup):
        assert(isinstance(modeGroup,autosar.mode.ModeGroup))
        lines=[]
        ws = modeGroup.rootWS()
        assert(ws is not None)
        lines.append('<%s>'%modeGroup.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%modeGroup.name,1))
        typeElem = ws.find(modeGroup.typeRef)
        if (typeElem is None):
            raise ValueError("invalid type reference: '%s'"%modeGroup.typeRef)
        else:
            lines.append(self.indent('<TYPE-TREF DEST="%s">%s</TYPE-TREF>'%(typeElem.tag(self.version),typeElem.ref),1))
        lines.append('</%s>'%modeGroup.tag(self.version))
        return lines

    def writeSoftwareAddressMethodXML(self, addressMethod):
        assert(isinstance(addressMethod,autosar.element.SoftwareAddressMethod))
        lines=[]
        lines.append('<%s>'%addressMethod.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%addressMethod.name,1))
        lines.append('</%s>'%addressMethod.tag(self.version))
        return lines

    def writeModeSwitchInterfaceXML(self, portInterface):
        assert(isinstance(portInterface, autosar.portinterface.ModeSwitchInterface))
        lines=[]
        lines.append('<%s>'%portInterface.tag(self.version))
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%portInterface.name,1))
        if portInterface.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(portInterface.adminData),1))
        lines.append(self.indent('<IS-SERVICE>%s</IS-SERVICE>'%self.toBooleanStr(portInterface.isService),1))
        lines.extend(self.indent(self.writeModeGroupXML(portInterface.modeGroup),1))
        lines.append('</%s>'%portInterface.tag(self.version))
        return lines

    def writeInvalidationPolicyXML(self, ws, invalidationPolicy):
        assert(isinstance(invalidationPolicy, autosar.portinterface.InvalidationPolicy))
        lines=[]
        lines.append('<%s>'%invalidationPolicy.tag(self.version))
        dataElement = ws.find(invalidationPolicy.dataElementRef, role="PortInterface")
        if (dataElement is None):
            raise ValueError("invalid type reference: '%s'"%invalidationPolicy.dataElementRef)
        else:
            lines.append(self.indent('<DATA-ELEMENT-REF DEST="%s">%s</DATA-ELEMENT-REF>'%(dataElement.tag(self.version), dataElement.ref),1))
            lines.append(self.indent('<HANDLE-INVALID>%s</HANDLE-INVALID>'%invalidationPolicy.handleInvalid,1))
        lines.append('</%s>'%invalidationPolicy.tag(self.version))
        return lines

    def writeNvDataInterfaceXML(self, portInterface):
        assert(isinstance(portInterface,autosar.portinterface.NvDataInterface))
        ws = portInterface.rootWS()
        lines=[]
        lines.append('<NV-DATA-INTERFACE>')
        lines.append(self.indent('<SHORT-NAME>%s</SHORT-NAME>'%portInterface.name,1))
        if portInterface.adminData is not None:
            lines.extend(self.indent(self.writeAdminDataXML(portInterface.adminData),1))
        lines.append(self.indent('<IS-SERVICE>%s</IS-SERVICE>'%self.toBooleanStr(portInterface.isService),1))
        if (self.version >= 4.0) and (portInterface.serviceKind is not None):
            lines.append(self.indent('<SERVICE-KIND>%s</SERVICE-KIND>'%portInterface.serviceKind,1))
        if len(portInterface.nvDatas)>0:
            lines.append(self.indent('<NV-DATAS>',1))
            for elem in portInterface.nvDatas:
                lines.extend(self.indent(self.writeDataElementXML(elem),2))
            lines.append(self.indent('</NV-DATAS>',1))
        else:
            lines.append(self.indent('<NV-DATAS/>',1))
        lines.append('</NV-DATA-INTERFACE>')
        return lines

class CodePortInterfaceWriter(ElementWriter):
    def __init__(self,version, patch):
        super().__init__(version, patch)

        if self.version >= 3.0 and self.version < 4.0:
            self.switcher = {
                             'SoftwareAddressMethod': self.writeSoftwareAddressMethodCode,
                             'ModeDeclarationGroup': self.writeModeDeclarationGroupCode,
                             'SenderReceiverInterface': self.writeSenderReceiverInterfaceCode,
                             'ParameterInterface': self.writeParameterInterfaceCode,
                             'ClientServerInterface': self.writeClientServerInterfaceCode,
            }
        elif self.version >= 4.0:
            self.switcher = {
            }
        else:
            switch.keys = {}

    def getSupportedXML(self):
        return []

    def getSupportedCode(self):
        return self.switcher.keys()

    def writeElementXML(self, elem):
        raise NotImplementedError('writeElementXML')

    def writeElementCode(self, elem, localvars):
        codeWriteFunc = self.switcher.get(type(elem).__name__)
        if codeWriteFunc is not None:
            return codeWriteFunc(elem, localvars)
        else:
            return None

    def writeSenderReceiverInterfaceCode(self, portInterface, localvars):
        assert(isinstance(portInterface,autosar.portinterface.SenderReceiverInterface))
        lines=[]
        params=['"%s"'%portInterface.name]
        if len(portInterface.dataElements)>1:
            raise NotImplementedError('more than one data element in an interface not yet supported')
        elif len(portInterface.dataElements)==1:
            params.append(self.writeDataElementCode(portInterface.dataElements[0], localvars))

        if portInterface.modeGroups is not None:
            if len(portInterface.modeGroups)>1:
                raise NotImplementedError("more then one modegroup not yet supported")
            elif len(portInterface.modeGroups)==1:
                params.append('modeGroups='+self.writeModeGroupCode(portInterface.modeGroups[0], localvars))

        if portInterface.isService:
            params.append('isService=True')
        if portInterface.adminData is not None:
            param = self.writeAdminDataCode(portInterface.adminData, localvars)
            assert(len(param)>0)
            params.append('adminData='+param)
        lines.append('package.createSenderReceiverInterface(%s)'%(', '.join(params)))
        return lines

    def writeDataElementCode(self, elem, localvars):
        ws = elem.rootWS()
        assert(ws is not None)
        dataType = ws.find(elem.typeRef, role="DataType")
        if dataType is None:
            raise ValueError('invalid reference: '+elem.typeRef)
        #name
        params=[repr(elem.name)]
        #typeRef
        if ws.roles['DataType'] is not None:
            params.append(repr(dataType.name)) #use name only
        else:
            params.append(repr(dataType.ref)) #use full reference
        if elem.isQueued:
            params.append('True')
        if elem.swAddressMethodRef is not None:
            params.append('softwareAddressMethodRef="%s"'%elem.swAddressMethodRef)
        if elem.adminData is not None:
            param = self.writeAdminDataCode(elem.adminData, localvars)
            assert(len(param)>0)
            params.append('adminData='+param)
        return 'autosar.DataElement(%s)'%(', '.join(params))

    def writeParameterCode(self, elem, localvars):
        ws = elem.rootWS()
        assert(ws is not None)
        dataType = ws.find(elem.typeRef, role="DataType")
        if dataType is None:
            raise ValueError('invalid reference: '+elem.typeRef)
        #name
        params=[repr(elem.name)]
        #typeRef
        if ws.roles['DataType'] is not None:
            params.append(repr(dataType.name)) #use name only
        else:
            params.append(repr(dataType.ref)) #use full reference
        if elem.swAddressMethodRef is not None:
            params.append('swAddressMethodRef="%s"'%elem.swAddressMethodRef)
        if elem.swCalibrationAccess is not None:
            params.append('swCalibrationAccess="%s"'%elem.swCalibrationAccess)
        if elem.adminData is not None:
            param = self.writeAdminDataCode(elem.adminData, localvars)
            assert(len(param)>0)
            params.append('adminData='+param)
        return 'autosar.Parameter(%s)'%(', '.join(params))

    def writeModeGroupCode(self, modeGroup, localvars):
        ws = modeGroup.rootWS()
        assert(ws is not None)
        dataType = ws.find(modeGroup.typeRef, role="ModeDclrGroup")
        if dataType is None:
            raise ValueError('invalid reference: '+modeGroup.typeRef)
        params=['"%s"'%modeGroup.name, '"%s"'%dataType.ref]
        return 'autosar.ModeGroup(%s)'%(', '.join(params))



    def writeParameterInterfaceCode(self, portInterface, localvars):
        assert(isinstance(portInterface,autosar.portinterface.ParameterInterface))
        lines=[]
        params=['"%s"'%portInterface.name]
        if len(portInterface.elements)==1:
            code=self.writeParameterCode(portInterface.elements[0], localvars)
            if (portInterface.elements[0].adminData is not None) or len(portInterface.elements[0].swAddrMethodRefList)>0:
                #this is going to be a long line, create separate dataElement variable
                lines.append('parameter=%s'%code)
                params.append('parameter')
            else:
                params.append(code)
        elif len(portInterface.elements)>1:
            raise NotImplementedError('more than one data element in an interface not yet supported')
        if portInterface.isService:
            params.append('isService=True')
        if portInterface.adminData is not None:
            param = self.writeAdminDataCode(portInterface.adminData, localvars)
            assert(len(param)>0)
            params.append('adminData='+param)
        lines.append('package.createParameterInterface(%s)'%(', '.join(params)))
        return lines

    def writeClientServerInterfaceCode(self, portInterface, localvars):
        lines=[]
        ws = portInterface.rootWS()
        assert(ws is not None)
        params=['"%s"'%portInterface.name]
        params2=[]
        for operation in portInterface.operations:
            params2.append('"%s"'%operation.name)
        if len(params2)>3:
            lines.extend(self.writeListCode("operationsList",params2))
            params.append('operationsList')
        else:
            params.append('['+', '.join(params2)+']')
        params2=[]
        for error in portInterface.applicationErrors:
            params2.append('autosar.ApplicationError("%s", %d)'%(error.name,error.errorCode))
        if len(params2)>1:
            lines.extend(self.writeListCode("errorsList",params2))
            params.append('errorsList')
        elif len(params2)==1:
            params.append(params2[0])

        if portInterface.isService:
            params.append('isService=True')
        if portInterface.adminData is not None:
            param = self.writeAdminDataCode(portInterface.adminData, localvars)
            assert(len(param)>0)
            params.append('adminData='+param)
        lines.append('portInterface=package.createClientServerInterface(%s)'%(', '.join(params)))
        localvars['portInterface']=portInterface
        for operation in portInterface.operations:
            for argument in operation.arguments:
                methodLookup={"IN": "createInArgument", "OUT": "createOutArgument", "INOUT": "createInOutArgument"}
                dataType = ws.find(argument.typeRef)
                if dataType is None:
                    raise ValueError("invalid reference: "+argument.typeRef)
                lines.append('portInterface["%s"].%s("%s", "%s")'%(operation.name, methodLookup[argument.direction], argument.name, dataType.name))
            params=[]
            if len(operation.errorRefs)>0:
                for ref in operation.errorRefs:
                    error = ws.find(ref)
                    if error is None:
                        raise ValueError("invalid reference: "+ref)
                    params.append('"%s"'%error.name)
                lines.append('portInterface["%s"].possibleErrors = %s'%(operation.name, ', '.join(params)))
            desc,descAttr=self.writeDescCode(operation)
            if desc is not None:
                lines.append('portInterface["%s"].desc = "%s"'%(operation.name, desc))
            if descAttr is not None:
                lines.append('portInterface["%s"].descAttr = "%s"'%(operation.name, descAttr))
        return lines

    def writeSoftwareAddressMethodCode(self, method, localvars):
        lines=[]
        lines.append('%s.createSoftwareAddressMethod("%s")'%('package', method.name))
        return lines

    def writeModeDeclarationGroupCode(self, declarationGroup, localvars):
        lines=[]
        params=['"%s"'%declarationGroup.name]
        params2=[]
        for item in declarationGroup.modeDeclarations:
            params2.append('"%s"'%item.name)
        if len(params2)>6:
            lines.extend(self.writeListCode("modeDeclarationsList",params2))
            params.append('modeDeclarationsList')
        else:
            params.append('['+', '.join(params2)+']')
        assert(declarationGroup.initialModeRef is not None)
        tmp=autosar.base.splitRef(declarationGroup.initialModeRef)
        params.append('"%s"'%tmp[-1])

        if declarationGroup.adminData is not None:
            param = self.writeAdminDataCode(declarationGroup.adminData, localvars)
            assert(len(param)>0)
            params.append('adminData='+param)

        lines.append('package.createModeDeclarationGroup(%s)'%(', '.join(params)))
        return lines



import sys
from autosar.base import splitRef, hasAdminData, parseAdminDataNode
import autosar.component
from autosar.parser.behavior_parser import BehaviorParser
from autosar.parser.parser_base import ElementParser
from autosar.parser.constant_parser import ConstantParser

def _getDataElemNameFromComSpec(xmlElem,portInterfaceRef):
    if xmlElem.find('./DATA-ELEMENT-REF') is not None:
        dataElemRef = splitRef(xmlElem.find('./DATA-ELEMENT-REF').text)
        assert(dataElemRef is not None)
        dataElemName = dataElemRef.pop()
        tmp = '/'+'/'.join(dataElemRef)
        if portInterfaceRef == tmp:
            return dataElemName
    return None

def _getOperationNameFromComSpec(xmlElem,portInterfaceRef):
    xmlOperation=xmlElem.find('./OPERATION-REF')
    if xmlOperation is not None:
        operationRef = splitRef(xmlOperation.text)
        assert(operationRef is not None)
        operationName = operationRef.pop()
        tmp = '/'+'/'.join(operationRef)
        if portInterfaceRef == tmp:
            return operationName
    return None

def _getParameterNameFromComSpec(xmlElem,portInterfaceRef):
    if xmlElem.tag == 'PARAMETER-REF':
        parameterRef = splitRef(xmlElem.text)
        assert(parameterRef is not None)
        name = parameterRef.pop()
        tmp = '/'+'/'.join(parameterRef)
        if portInterfaceRef == tmp:
            return name
    return None

def _getVariableNameFromComSpec(xmlElem,portInterfaceRef):
    if xmlElem.tag == 'VARIABLE-REF':
        variableRef = splitRef(xmlElem.text)
        assert(variableRef is not None)
        name = variableRef.pop()
        tmp = '/'+'/'.join(variableRef)
        if portInterfaceRef == tmp:
            return name
    return None

class ComponentTypeParser(ElementParser):
    """
    ComponentType parser
    """

    def __init__(self,version=3.0):
        super().__init__(version)
        if self.version >=4.0:
            self.behavior_parser = BehaviorParser(version)
            self.constant_parser = ConstantParser(version)

        if self.version >= 3.0 and self.version < 4.0:
            self.switcher = { 'APPLICATION-SOFTWARE-COMPONENT-TYPE': self.parseSoftwareComponent,
                              'COMPLEX-DEVICE-DRIVER-COMPONENT-TYPE': self.parseSoftwareComponent,
                              'COMPOSITION-TYPE': self.parseCompositionType,
                              'CALPRM-COMPONENT-TYPE': self.parseSoftwareComponent,
                              'SERVICE-COMPONENT-TYPE': self.parseSoftwareComponent
                             }
        elif self.version >= 4.0:
            self.switcher = {
               'APPLICATION-SW-COMPONENT-TYPE': self.parseSoftwareComponent,
               'COMPLEX-DEVICE-DRIVER-SW-COMPONENT-TYPE': self.parseSoftwareComponent,
               'SERVICE-COMPONENT-TYPE': self.parseSoftwareComponent,
               'PARAMETER-SW-COMPONENT-TYPE': self.parseSoftwareComponent,
               'COMPOSITION-SW-COMPONENT-TYPE': self.parseCompositionType,
               'SENSOR-ACTUATOR-SW-COMPONENT-TYPE': self.parseSoftwareComponent,
               'SERVICE-SW-COMPONENT-TYPE': self.parseSoftwareComponent,
               'NV-BLOCK-SW-COMPONENT-TYPE': self.parseSoftwareComponent
            }

    def getSupportedTags(self):
        return self.switcher.keys()

    def parseElement(self, xmlElement, parent = None):
        parseFunc = self.switcher.get(xmlElement.tag)
        if parseFunc is not None:
            return parseFunc(xmlElement,parent)
        else:
            return None

    def parseSoftwareComponent(self, xmlRoot, parent=None):
        componentType=None
        handledTags = ['SHORT-NAME']
        if xmlRoot.tag=='APPLICATION-SOFTWARE-COMPONENT-TYPE': #for AUTOSAR 3.x
            componentType = autosar.component.ApplicationSoftwareComponent(self.parseTextNode(xmlRoot.find('SHORT-NAME')),parent)
        elif (xmlRoot.tag=='COMPLEX-DEVICE-DRIVER-COMPONENT-TYPE') or (xmlRoot.tag=='COMPLEX-DEVICE-DRIVER-SW-COMPONENT-TYPE'):
            componentType = autosar.component.ComplexDeviceDriverComponent(self.parseTextNode(xmlRoot.find('SHORT-NAME')),parent)
        elif xmlRoot.tag == 'APPLICATION-SW-COMPONENT-TYPE': #for AUTOSAR 4.x
            componentType = autosar.component.ApplicationSoftwareComponent(self.parseTextNode(xmlRoot.find('SHORT-NAME')),parent)
        elif (xmlRoot.tag == 'SERVICE-COMPONENT-TYPE') or (xmlRoot.tag == 'SERVICE-SW-COMPONENT-TYPE'):
            componentType = autosar.component.ServiceComponent(self.parseTextNode(xmlRoot.find('SHORT-NAME')),parent)
        elif xmlRoot.tag == 'CALPRM-COMPONENT-TYPE': #for AUTOSAR 3.x
            componentType = autosar.component.ParameterComponent(self.parseTextNode(xmlRoot.find('SHORT-NAME')),parent)
        elif xmlRoot.tag == 'PARAMETER-SW-COMPONENT-TYPE': #for AUTOSAR 4.x
            componentType = autosar.component.ParameterComponent(self.parseTextNode(xmlRoot.find('SHORT-NAME')),parent)
        elif xmlRoot.tag == 'SENSOR-ACTUATOR-SW-COMPONENT-TYPE': #for AUTOSAR 4.x
            componentType = autosar.component.SensorActuatorComponent(self.parseTextNode(xmlRoot.find('SHORT-NAME')),parent)
        elif xmlRoot.tag == 'NV-BLOCK-SW-COMPONENT-TYPE': #for AUTOSAR 4.x
            componentType = autosar.component.NvBlockComponent(self.parseTextNode(xmlRoot.find('SHORT-NAME')),parent)
        else:
            raise NotImplementedError(xmlRoot.tag)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag not in handledTags:
                if xmlElem.tag == 'ADMIN-DATA':
                    componentType.adminData = self.parseAdminDataNode(xmlElem)
                elif xmlElem.tag == 'DESC':
                    pass #Implement later
                elif xmlElem.tag == 'PORTS':
                    self.parseComponentPorts(componentType,xmlRoot)
                elif xmlElem.tag == 'INTERNAL-BEHAVIORS':
                    behaviors = xmlElem.findall('./SWC-INTERNAL-BEHAVIOR')
                    if len(behaviors)>1:
                        raise ValueError('%s: an SWC cannot have multiple internal behaviors'%(componentType))
                    elif len(behaviors) == 1:
                        componentType.behavior = self.behavior_parser.parseSWCInternalBehavior(behaviors[0], componentType)
                elif xmlElem.tag == 'NV-BLOCK-DESCRIPTORS' and isinstance(componentType, autosar.component.NvBlockComponent):
                    for descriptorXml in xmlElem.findall('./NV-BLOCK-DESCRIPTOR'):
                        descriptor = self.behavior_parser.parseNvBlockSWCnvBlockDescriptor(descriptorXml, componentType)
                        componentType.nvBlockDescriptors.append(descriptor)
                elif xmlElem.tag == 'CATEGORY':
                    componentType.category = self.parseTextNode(xmlElem)
                else:
                    print('Unhandled tag: '+xmlElem.tag, file=sys.stderr)
        return componentType

    def parseComponentPorts(self, componentType, xmlRoot):
        xmlPorts=xmlRoot.find('PORTS')
        assert(xmlPorts is not None)
        for xmlPort in xmlPorts.findall('*'):
            if(xmlPort.tag == "R-PORT-PROTOTYPE"):
                portName = xmlPort.find('SHORT-NAME').text
                portInterfaceRef = self.parseTextNode(xmlPort.find('REQUIRED-INTERFACE-TREF'))
                port = autosar.port.RequirePort(portName, portInterfaceRef, autoCreateComSpec = False, parent=componentType)
                if hasAdminData(xmlPort):
                    port.adminData=parseAdminDataNode(xmlPort.find('ADMIN-DATA'))
                if xmlPort.findall('./REQUIRED-COM-SPECS') is not None:
                    for xmlItem in xmlPort.findall('./REQUIRED-COM-SPECS/*'):
                        if xmlItem.tag == 'CLIENT-COM-SPEC':
                            operationName=_getOperationNameFromComSpec(xmlItem,portInterfaceRef)
                            comspec=autosar.port.OperationComSpec(operationName)
                            port.comspec.append(comspec)
                        elif xmlItem.tag == 'UNQUEUED-RECEIVER-COM-SPEC' or xmlItem.tag == 'NONQUEUED-RECEIVER-COM-SPEC':
                            dataElemName = _getDataElemNameFromComSpec(xmlItem,portInterfaceRef)
                            comspec = autosar.port.DataElementComSpec(dataElemName)
                            if xmlItem.find('./ALIVE-TIMEOUT') is not None:
                                comspec.aliveTimeout = self.parseTextNode(xmlItem.find('./ALIVE-TIMEOUT'))
                            if self.version >= 4.0:
                                xmlElem = xmlItem.find('./INIT-VALUE')
                                if xmlElem != None:
                                    comspec.initValue, comspec.initValueRef = self._parseAr4InitValue(xmlElem)
                            else:
                                if xmlItem.find('./INIT-VALUE-REF') != None:
                                    comspec.initValueRef = self.parseTextNode(xmlItem.find('./INIT-VALUE-REF'))
                            port.comspec.append(comspec)
                        elif xmlItem.tag == 'QUEUED-RECEIVER-COM-SPEC':
                            dataElemName = _getDataElemNameFromComSpec(xmlItem,portInterfaceRef)
                            comspec = autosar.port.DataElementComSpec(dataElemName)
                            if xmlItem.find('./QUEUE-LENGTH') != None:
                                comspec.queueLength = self.parseTextNode(xmlItem.find('./QUEUE-LENGTH'))
                            port.comspec.append(comspec)
                        elif xmlItem.tag == 'MODE-SWITCH-RECEIVER-COM-SPEC':
                            comspec = self._parseModeSwitchReceiverComSpec(xmlItem)
                            port.comspec.append(comspec)
                        elif xmlItem.tag == 'PARAMETER-REQUIRE-COM-SPEC':
                            comspec = self._parseParameterComSpec(xmlItem, portInterfaceRef)
                            port.comspec.append(comspec)
                        elif xmlItem.tag == 'NV-REQUIRE-COM-SPEC':
                            comspec = self._parseNvRequireComSpec(xmlItem, portInterfaceRef)
                            assert(comspec is not None)
                            port.comspec.append(comspec)
                        else:
                            raise NotImplementedError(xmlItem.tag)
                componentType.requirePorts.append(port)
            elif(xmlPort.tag == 'P-PORT-PROTOTYPE'):
                portName = xmlPort.find('SHORT-NAME').text
                portInterfaceRef = self.parseTextNode(xmlPort.find('PROVIDED-INTERFACE-TREF'))
                port = autosar.port.ProvidePort(portName, portInterfaceRef, autoCreateComSpec = False, parent = componentType)
                if hasAdminData(xmlPort):
                    port.adminData=parseAdminDataNode(xmlPort.find('ADMIN-DATA'))
                if xmlPort.findall('./PROVIDED-COM-SPECS') is not None:
                    for xmlItem in xmlPort.findall('./PROVIDED-COM-SPECS/*'):
                        if xmlItem.tag == 'SERVER-COM-SPEC':
                            operationName=_getOperationNameFromComSpec(xmlItem,portInterfaceRef)
                            comspec=autosar.port.OperationComSpec(operationName)
                            comspec.queueLength=self.parseIntNode(xmlItem.find('QUEUE-LENGTH'))
                            port.comspec.append(comspec)
                        elif xmlItem.tag == 'UNQUEUED-SENDER-COM-SPEC' or xmlItem.tag == 'NONQUEUED-SENDER-COM-SPEC':
                            dataElemName = _getDataElemNameFromComSpec(xmlItem,portInterfaceRef)
                            comspec = autosar.port.DataElementComSpec(dataElemName)
                            if self.version >= 4.0:
                                xmlElem = xmlItem.find('./INIT-VALUE')
                                if xmlElem != None:
                                    comspec.initValue, comspec.initValueRef = self._parseAr4InitValue(xmlElem)
                            else:
                                if xmlItem.find('./INIT-VALUE-REF') is not None:
                                    comspec.initValueRef = self.parseTextNode(xmlItem.find('./INIT-VALUE-REF'))
                            if xmlItem.find('./CAN-INVALIDATE') != None:
                                comspec.canInvalidate = True if self.parseTextNode(xmlItem.find('./CAN-INVALIDATE'))=='true' else False
                            port.comspec.append(comspec)
                        elif xmlItem.tag == 'QUEUED-SENDER-COM-SPEC':
                            dataElemName = _getDataElemNameFromComSpec(xmlItem,portInterfaceRef)
                            comspec = autosar.port.DataElementComSpec(dataElemName)
                            assert(comspec is not None)
                            port.comspec.append(comspec)
                        elif xmlItem.tag == 'PARAMETER-PROVIDE-COM-SPEC':
                            comspec = self._parseParameterComSpec(xmlItem, portInterfaceRef)
                            assert(comspec is not None)
                            port.comspec.append(comspec)
                        elif xmlItem.tag == 'MODE-SWITCH-SENDER-COM-SPEC':
                            comspec = self._parseModeSwitchSenderComSpec(xmlItem)
                            assert(comspec is not None)
                            port.comspec.append(comspec)
                        elif xmlItem.tag == 'NV-PROVIDE-COM-SPEC':
                            comspec = self._parseNvProvideComSpec(xmlItem, portInterfaceRef)
                            assert(comspec is not None)
                            port.comspec.append(comspec)
                        else:
                            raise NotImplementedError(xmlItem.tag)
                componentType.providePorts.append(port)

    def parseCompositionType(self, xmlRoot, parent=None):
        """
        parses COMPOSITION-TYPE
        """
        assert (xmlRoot.tag=='COMPOSITION-TYPE') or (xmlRoot.tag=='COMPOSITION-SW-COMPONENT-TYPE')
        dataTypeMappingRefs = None
        swc=autosar.component.CompositionComponent(self.parseTextNode(xmlRoot.find('SHORT-NAME')),parent)
        self.push()
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag=='SHORT-NAME':
                continue
            elif xmlElem.tag=='PORTS':
                self.parseComponentPorts(swc,xmlRoot)
            elif xmlElem.tag=='COMPONENTS':
                self.parseComponents(xmlElem,swc)
            elif xmlElem.tag=='CONNECTORS':
                if self.version >= 4.0:
                    self.parseConnectorsV4(xmlElem,swc)
                else:
                    self.parseConnectorsV3(xmlElem,swc)
            elif xmlElem.tag == 'DATA-TYPE-MAPPING-REFS':
                dataTypeMappingRefs = []
                for xmlChild in xmlElem.findall('./*'):
                    if xmlChild.tag == 'DATA-TYPE-MAPPING-REF':
                        tmp = self.parseTextNode(xmlChild)
                        assert(tmp is not None)
                        dataTypeMappingRefs.append(tmp)
            elif xmlElem.tag=='PORT-GROUPS':
                print("[ComponentParser] unhandled: %s"%xmlElem.tag)
            else:
                self.defaultHandler(xmlElem)
        if dataTypeMappingRefs is not None:
            swc.dataTypeMappingRefs = dataTypeMappingRefs
        self.pop(swc)
        return swc

    def parseComponents(self,xmlRoot,parent):
        """
        parses <COMPONENTS>
        """
        assert(xmlRoot.tag=='COMPONENTS')
        for elem in xmlRoot.findall('./*'):
            componentTag = 'SW-COMPONENT-PROTOTYPE' if self.version >= 4.0 else 'COMPONENT-PROTOTYPE'
            if elem.tag==componentTag:
                name=self.parseTextNode(elem.find('SHORT-NAME'))
                typeRef=self.parseTextNode(elem.find('TYPE-TREF'))
                parent.components.append(autosar.component.ComponentPrototype(name,typeRef,parent))
            else:
                raise NotImplementedError(elem.tag)

    def parseConnectorsV3(self,xmlRoot,parent=None):
        """
        parses <CONNECTORS> (AUTOSAR 3)
        """
        assert(xmlRoot.tag=='CONNECTORS')
        for elem in xmlRoot.findall('./*'):
            if elem.tag=='ASSEMBLY-CONNECTOR-PROTOTYPE':
                name=self.parseTextNode(elem.find('SHORT-NAME'))
                providerComponentRef=self.parseTextNode(elem.find('./PROVIDER-IREF/COMPONENT-PROTOTYPE-REF'))
                providerPortRef=self.parseTextNode(elem.find('./PROVIDER-IREF/P-PORT-PROTOTYPE-REF'))
                requesterComponentRef=self.parseTextNode(elem.find('./REQUESTER-IREF/COMPONENT-PROTOTYPE-REF'))
                requesterPortRef=self.parseTextNode(elem.find('./REQUESTER-IREF/R-PORT-PROTOTYPE-REF'))
                parent.assemblyConnectors.append(autosar.component.AssemblyConnector(name, autosar.component.ProviderInstanceRef(providerComponentRef,providerPortRef), autosar.component.RequesterInstanceRef(requesterComponentRef,requesterPortRef)))
            elif elem.tag=='DELEGATION-CONNECTOR-PROTOTYPE':
                name=self.parseTextNode(elem.find('SHORT-NAME'))
                innerComponentRef=self.parseTextNode(elem.find('./INNER-PORT-IREF/COMPONENT-PROTOTYPE-REF'))
                innerPortRef=self.parseTextNode(elem.find('./INNER-PORT-IREF/PORT-PROTOTYPE-REF'))
                outerPortRef=self.parseTextNode(elem.find('./OUTER-PORT-REF'))
                parent.delegationConnectors.append(autosar.component.DelegationConnector(name, autosar.component.InnerPortInstanceRef(innerComponentRef,innerPortRef), autosar.component.OuterPortRef(outerPortRef)))
            else:
                raise NotImplementedError(elem.tag)

    def parseConnectorsV4(self,xmlRoot,parent=None):
        """
        parses <CONNECTORS> (AUTOSAR 4)
        """
        assert(xmlRoot.tag=='CONNECTORS')
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag=='ASSEMBLY-SW-CONNECTOR':
                name=self.parseTextNode(xmlElem.find('SHORT-NAME'))
                for xmlChild in xmlElem.findall('./*'):
                    if xmlChild.tag == 'SHORT-NAME':
                        continue
                    elif xmlChild.tag == 'PROVIDER-IREF':
                        providerComponentRef=self.parseTextNode(xmlChild.find('./CONTEXT-COMPONENT-REF'))
                        providerPortRef=self.parseTextNode(xmlChild.find('./TARGET-P-PORT-REF'))
                    elif xmlChild.tag == 'REQUESTER-IREF':
                        requesterComponentRef=self.parseTextNode(xmlChild.find('./CONTEXT-COMPONENT-REF'))
                        requesterPortRef=self.parseTextNode(xmlChild.find('./TARGET-R-PORT-REF'))
                    else:
                        raise NotImplementedError(xmlChild.tag)
                if providerComponentRef is None:
                    raise RuntimeError('PROVIDER-IREF/CONTEXT-COMPONENT-REF is missing: item=%s'%name)
                if providerComponentRef is None:
                    raise RuntimeError('PROVIDER-IREF/TARGET-P-PORT-REF is missing: item=%s'%name)
                if requesterComponentRef is None:
                    raise RuntimeError('REQUESTER-IREF/CONTEXT-COMPONENT-REF is missing: item=%s'%name)
                if requesterPortRef is None:
                    raise RuntimeError('REQUESTER-IREF/TARGET-R-PORT-REF is missing: item=%s'%name)

                parent.assemblyConnectors.append(autosar.component.AssemblyConnector(name, autosar.component.ProviderInstanceRef(providerComponentRef,providerPortRef), autosar.component.RequesterInstanceRef(requesterComponentRef,requesterPortRef)))
            elif xmlElem.tag=='DELEGATION-SW-CONNECTOR':
                name=self.parseTextNode(xmlElem.find('SHORT-NAME'))
                for xmlChild in xmlElem.findall('./INNER-PORT-IREF/*'):
                    if xmlChild.tag == 'R-PORT-IN-COMPOSITION-INSTANCE-REF':
                        innerComponentRef=self.parseTextNode(xmlChild.find('./CONTEXT-COMPONENT-REF'))
                        innerPortRef=self.parseTextNode(xmlChild.find('./TARGET-R-PORT-REF'))
                    elif xmlChild.tag == 'P-PORT-IN-COMPOSITION-INSTANCE-REF':
                        innerComponentRef=self.parseTextNode(xmlChild.find('./CONTEXT-COMPONENT-REF'))
                        innerPortRef=self.parseTextNode(xmlChild.find('./TARGET-P-PORT-REF'))
                    else:
                        raise NotImplementedError(xmlChild.tag)
                outerPortRef=self.parseTextNode(xmlElem.find('./OUTER-PORT-REF'))
                parent.delegationConnectors.append(autosar.component.DelegationConnector(name, autosar.component.InnerPortInstanceRef(innerComponentRef,innerPortRef), autosar.component.OuterPortRef(outerPortRef)))
            else:
                raise NotImplementedError(xmlElem.tag)

    def _parseModeSwitchReceiverComSpec(self, xmlRoot):
        (enhancedMode, supportAsync, modeGroupRef) = (None, None, None)
        assert(xmlRoot.tag == 'MODE-SWITCH-RECEIVER-COM-SPEC')
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'ENHANCED-MODE-API':
                enhancedMode = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'SUPPORTS-ASYNCHRONOUS-MODE-SWITCH':
                supportAsync = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'MODE-GROUP-REF':
                modeGroupRef = self.parseTextNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        return autosar.port.ModeSwitchComSpec(None, enhancedMode, supportAsync, modeGroupRef = modeGroupRef)

    def _parseModeSwitchSenderComSpec(self, xmlRoot):
        (enhancedMode, queueLength, modeSwitchAckTimeout, modeGroupRef) = (None, None, None, None)
        assert(xmlRoot.tag == 'MODE-SWITCH-SENDER-COM-SPEC')
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'ENHANCED-MODE-API':
                enhancedMode = self.parseBooleanNode(xmlElem)
            elif xmlElem.tag == 'MODE-GROUP-REF':
                modeGroupRef = self.parseTextNode(xmlElem)
            elif xmlElem.tag == 'MODE-SWITCHED-ACK':
                for xmlChild in xmlElem.findall('./*'):
                    if xmlChild.tag == 'TIMEOUT':
                        tmp = self.parseFloatNode(xmlChild)
                        if tmp is not None:
                            modeSwitchAckTimeout = int(tmp*1000) #We use milliseconds in our internal model
            elif xmlElem.tag == 'QUEUE-LENGTH':
                queueLength = self.parseIntNode(xmlElem)
            else:
                raise NotImplementedError(xmlElem.tag)
        return autosar.port.ModeSwitchComSpec(None, enhancedMode, None, queueLength, modeSwitchAckTimeout, modeGroupRef)

    def _parseParameterComSpec(self, xmlRoot, portInterfaceRef):
        (initValue, name) = (None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'INIT-VALUE':
                initValue, initValueRef = self._parseAr4InitValue(xmlElem)
                if initValueRef is not None:
                    raise NotImplementedError('CONSTANT-REFERENCE')
            elif xmlElem.tag == 'PARAMETER-REF':
                name = _getParameterNameFromComSpec(xmlElem, portInterfaceRef)
            else:
                raise NotImplementedError(xmlElem.tag)
        if (name is not None):
            return autosar.port.ParameterComSpec(name, initValue)
        else:
            raise RuntimeError('PARAMETER-REQUIRE-COM-SPEC must have a PARAMETER-REF')

    def _parseNvProvideComSpec(self, xmlRoot, portInterfaceRef):
        (romBlockInitValue, romBlockInitValueRef, ramBlockInitValue, ramBlockInitValueRef, name) = (None, None, None, None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'RAM-BLOCK-INIT-VALUE':
                ramBlockInitValue, ramBlockInitValueRef = self._parseAr4InitValue(xmlElem)
            elif xmlElem.tag == 'ROM-BLOCK-INIT-VALUE':
                romBlockInitValue, romBlockInitValueRef = self._parseAr4InitValue(xmlElem)
            elif xmlElem.tag == 'VARIABLE-REF':
                name = _getVariableNameFromComSpec(xmlElem, portInterfaceRef)
            else:
                raise NotImplementedError(xmlElem.tag)
        if (name is not None):
            return autosar.port.NvProvideComSpec(name, ramBlockInitValue=ramBlockInitValue,
                                                    ramBlockInitValueRef=ramBlockInitValueRef,
                                                    romBlockInitValue=romBlockInitValue,
                                                    romBlockInitValueRef=romBlockInitValueRef)
        else:
            raise RuntimeError('NV-PROVIDE-COM-SPEC must have a VARIABLE-REF')

    def _parseNvRequireComSpec(self, xmlRoot, portInterfaceRef):
        (initValue, initValueRef, name) = (None, None, None)
        for xmlElem in xmlRoot.findall('./*'):
            if xmlElem.tag == 'INIT-VALUE':
                initValue, initValueRef = self._parseAr4InitValue(xmlElem)
            elif xmlElem.tag == 'VARIABLE-REF':
                name = _getVariableNameFromComSpec(xmlElem, portInterfaceRef)
            else:
                raise NotImplementedError(xmlElem.tag)
        if (name is not None):
            return autosar.port.NvRequireComSpec(name, initValue=initValue,
                                                    initValueRef=initValueRef)
        else:
            raise RuntimeError('NV-PROVIDE-COM-SPEC must have a VARIABLE-REF')

    def _parseAr4InitValue(self, xmlElem):
        (initValue, initValueRef) = (None, None)
        for xmlChild in xmlElem.findall('./*'):
            if xmlChild.tag == 'CONSTANT-REFERENCE':
                initValueRef = self.parseTextNode(xmlChild.find('./CONSTANT-REF'))
            else:
                values = self.constant_parser.parseValueV4(xmlElem, None)
                if len(values) != 1:
                    raise ValueError('{0} cannot cannot contain multiple elements'.format(xmlElem.tag))
                initValue = values[0]
        return (initValue, initValueRef)



from setuptools import setup


def readme():
    with open('README.md') as f:
        return f.read()

setup(name='autosar',
      version='0.4.4',
      description='A set of Python modules for working with AUTOSAR XML files',
      long_description=readme(),
      long_description_content_type='text/markdown',
      classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.6',
        'Topic :: Text Processing :: Markup :: XML',
      ],
      url='http://github.com/cogu/autosar',
      project_urls={
        'Documentation': 'https://autosar.readthedocs.io/'
      },
      author='Conny Gustafsson',
      author_email='congus8@gmail.com',
      license='MIT',
	    install_requires=[
        'cfile==0.2.0',
      ],
      packages=['autosar','autosar.parser','autosar.writer','autosar.rte', 'autosar.bsw', 'autosar.util'],
	    zip_safe=False,
	    test_suite='tests.my_test_suite')



from autosar.writer.writer_base import BaseWriter, ElementWriter
from autosar.base import applyFilter
import collections.abc
import autosar.behavior
import autosar.component

class PackageWriter(BaseWriter):
    def __init__(self, version, patch):
        super().__init__(version, patch)
        self.registeredWriters={}
        self.xmlSwitcher={}
        self.codeSwitcher={}

    def registerElementWriter(self, elementWriter):
        """
        registers a new element writer into this package writer
        """
        assert(isinstance(elementWriter, ElementWriter))
        writerName = type(elementWriter).__name__
        if writerName not in self.registeredWriters:
            list_or_generator = elementWriter.getSupportedXML()
            if list_or_generator is not None:
                if isinstance(list_or_generator, str):
                    list_or_generator = [list_or_generator]
                for elementName in list_or_generator:
                    self.xmlSwitcher[elementName] = elementWriter
            list_or_generator = elementWriter.getSupportedCode()
            if list_or_generator is not None:
                if isinstance(list_or_generator, str):
                    list_or_generator = [list_or_generator]
                for elementName in list_or_generator:
                    self.codeSwitcher[elementName] = elementWriter
            self.registeredWriters[writerName] = elementWriter

    def toXML(self, package, filters, ignore):
        lines=[]
        lines.extend(self.beginPackage(package.name))
        if len(package.elements)>0:
            lines.append(self.indent("<ELEMENTS>",1))
            for elem in package.elements:
                elemRef = elem.ref
                ignoreElem=True if (isinstance(ignore, collections.abc.Iterable) and elemRef in ignore) else False
                #if SWC was ignored by user, also ignore its InternalBehavior and SwcImplementation elements in case they are in the same package
                if not ignoreElem and isinstance(elem, autosar.behavior.InternalBehavior):
                    if (isinstance(ignore, collections.abc.Iterable) and elem.componentRef in ignore):
                        ignoreElem = True
                if not ignoreElem and isinstance(elem, autosar.component.SwcImplementation):
                    behavior = package.rootWS().find(elem.behaviorRef)
                    if behavior is not None:
                        if (isinstance(ignore, collections.abc.Iterable) and behavior.componentRef in ignore):
                            ignoreElem = True
                if not ignoreElem and applyFilter(elemRef, filters):
                    elementName = elem.__class__.__name__
                    elementWriter = self.xmlSwitcher.get(elementName)
                    if elementWriter is not None:
                        result = elementWriter.writeElementXML(elem)
                        if result is None:
                            print("[PackageWriter] No return value: %s"%elementName)
                            continue
                        else:
                            lines.extend(self.indent(result,2))
                    else:
                        package.unhandledWriter.add(elementName)
            lines.append(self.indent("</ELEMENTS>",1))
        else:
            if self.version<4.0:
                lines.append(self.indent("<ELEMENTS/>",1))
        if len(package.subPackages)>0:
            numPackets = 0
            if self.version >= 3.0 and self.version < 4.0:
                for subPackage in package.subPackages:
                    if applyFilter(subPackage.ref, filters):
                        if numPackets == 0:
                            lines.append(self.indent("<SUB-PACKAGES>",1))
                        lines.extend(self.indent(self.toXML(subPackage, filters, ignore),2))
                        numPackets += 1
                if numPackets > 0:
                    lines.append(self.indent("</SUB-PACKAGES>",1))
            elif self.version >= 4.0:
                for subPackage in package.subPackages:
                    if applyFilter(subPackage.ref, filters):
                        if numPackets == 0:
                            lines.append(self.indent("<AR-PACKAGES>",1))
                        lines.extend(self.indent(self.toXML(subPackage, filters, ignore),2))
                        numPackets += 1
                if numPackets > 0:
                    lines.append(self.indent("</AR-PACKAGES>",1))
        lines.extend(self.endPackage())
        return lines

    def toCode(self, package, filters, ignore, localvars, isTemplate):
        lines=[]
        if not isTemplate:
            if package.role is not None:
                lines.append('package=ws.createPackage("%s", role="%s")'%(package.name, package.role))
            else:
                lines.append('package=ws.createPackage("%s")'%(package.name))
            localvars['package']=package
            for subPackage in package.subPackages:
                if subPackage.role is not None:
                    lines.append('package.createSubPackage("%s", role="%s")'%(subPackage.name, subPackage.role))
                else:
                    lines.append('package.createSubPackage("%s")'%(subPackage.name))
        for elem in package.elements:
            elemRef = elem.ref
            ignoreElem=True if (isinstance(ignore, str) and ignore==elemRef) or (isinstance(ignore, collections.abc.Iterable) and elemRef in ignore) else False

            #if SWC was ignored by user, also ignore its InternalBehavior and SwcImplementation elements in case they are in the same package
            if not ignoreElem and isinstance(elem, autosar.behavior.InternalBehavior):
                if (isinstance(ignore, str) and ignore==elem.componentRef) or (isinstance(ignore, collections.abc.Iterable) and elem.componentRef in ignore): ignoreElem = True
            if not ignoreElem and isinstance(elem, autosar.component.SwcImplementation):
                behavior = package.rootWS().find(elem.behaviorRef)
                if behavior is not None:
                    if (isinstance(ignore, str) and ignore==behavior.componentRef) or (isinstance(ignore, collections.abc.Iterable) and behavior.componentRef in ignore): ignoreElem = True
            if not ignoreElem and applyFilter(elemRef, filters):
                elementName = elem.__class__.__name__
                elementWriter = self.codeSwitcher.get(elementName)
                if elementWriter is not None:
                    result = elementWriter.writeElementCode(elem, localvars)
                    if result is None:
                        print("[PackageWriter] No return value: %s"%elementName)
                        continue
                    else:
                        lines.extend(result)
                else:
                    package.unhandledWriter.add(elementName)
            else:
                pass
        return lines



import ntpath
import os
import sys
import xml.etree.ElementTree as ElementTree
import autosar

dvg_xml = """<?xml version="1.0" encoding="utf-8"?>
<DVG>
	<Version>1.0</Version>
	<Contents>
	</Contents>
</DVG>"""

class XMLWriterSimple:

    def __init__(self):
        self.indentChar='\t'

    def make_dirs(self, dest_dir):
        if dest_dir != '.' and dest_dir != '..':
            if not os.path.exists(dest_dir):
                os.makedirs(dest_dir)

    def indent(self, lines, amount):
        if isinstance(lines, str):
            return self.indentChar*amount+lines
        elif isinstance(lines, list):
            return [self.indentChar*amount+line for line in lines]
        else:
            raise ValueError('lines must be string or list')

class DcfProfile(XMLWriterSimple):
    """
    Container for DaVinci Developer Profile Settings
    """
    def __init__(self):
        super().__init__()
        self.ExtendedShortNameLength=True
        self.DataMapping_Compatibility=True
        self.DataMapping_Defines=True
        self.DataMapping_InitValues=True
        self.DataMapping_InvalidValues = True
        self.DataMapping_Length = True
        self.DataMapping_Scale = True
        self.DataMapping_SignalGroups = True
        self.DataMapping_Unit = True
        self.DataType_UseTerminatingNullByteForStrings = False
        self.PortCompat_CheckDelegationConnectorComSpec = True
        self.SoftwareDesign_PortCompatibility = False
        self.UnconnectedPorts_Calibration = False
        self.UnconnectedPorts_ClientPorts = True
        self.UnconnectedPorts_ReceiverPorts = True
        self.UnconnectedPorts_SenderPorts = True
        self.UnconnectedPorts_ServerPorts = True
        self.ShowCalibrationObjects = False
        self.UseGraphicEditors = True
        self.DefaultPackage_DataTypes = "DataTypes"
        self.DefaultPackage_Constants = "Constants"
        self.DefaultPackage_PortInterfaces = "PortInterfaces"
        self.DefaultPackage_ComponentTypes = "ComponentTypes"
        self.DefaultPackage_ModeDeclarationGroups = "ModeDclrGroups"
        self.DefaultPackage_BaseTypes = "DataTypes/BaseTypes"
        self.DefaultPackage_Units = "DataTypes/Units"
        self.DefaultPackage_CompuMethods = "DataTypes/CompuMethods"
        self.DefaultPackage_DataConstrs = "DataTypes/DataConstrs"
        self.DefaultPackage_DataTypeMappingSets = "ComponentTypes/DataTypeMappingSets"
        self.DefaultPackage_EndToEndProtections = "EndToEndProtections"
        self.DefaultPackage_ECUCompositions = "ECUCompositions"
        self.DefaultPackage_ECUCompositionTypes = "ECUCompositionTypes"
        self.DefaultPackage_ECUProjects = "Systems"
        self.DefaultPackage_Blueprints = "Blueprints/PortPrototypeBlueprints"
        self.DefaultPackage_BlueprintMappingSets = "Blueprints/BlueprintMappingSets"
        self.DefaultPackage_SwAddrMethods = "SwAddrMethod"

        self.sectionKeys = {'AUTOSAR_XML': set(['ExtendedShortNameLength']),
                            'CHECK_AND_SYNCHRONIZE': set(['DataMapping_Compatibility',
                                                         'DataMapping_Defines',
                                                         'DataMapping_InitValues',
                                                         'DataMapping_InvalidValues',
                                                         'DataMapping_Length',
                                                         'DataMapping_Scale',
                                                         'DataMapping_SignalGroups',
                                                         'DataMapping_Unit',
                                                         'DataType_UseTerminatingNullByteForStrings',
                                                         'PortCompat_CheckDelegationConnectorComSpec',
                                                         'SoftwareDesign_PortCompatibility',
                                                         'UnconnectedPorts_Calibration',
                                                         'UnconnectedPorts_ClientPorts',
                                                         'UnconnectedPorts_ReceiverPorts',
                                                         'UnconnectedPorts_SenderPorts',
                                                         'UnconnectedPorts_ServerPorts'
                                                         ]),
                            'DEFAULT_PACKAGES': set(['DefaultPackage_DataTypes',
                                                     'DefaultPackage_Constants',
                                                     'DefaultPackage_PortInterfaces',
                                                     'DefaultPackage_ComponentTypes',
                                                     'DefaultPackage_ModeDeclarationGroups',
                                                     'DefaultPackage_BaseTypes',
                                                     'DefaultPackage_Units',
                                                     'DefaultPackage_CompuMethods',
                                                     'DefaultPackage_DataConstrs',
                                                     'DefaultPackage_DataTypeMappingSets',
                                                     'DefaultPackage_EndToEndProtections',
                                                     'DefaultPackage_ECUCompositions',
                                                     'DefaultPackage_ECUCompositionTypes',
                                                     'DefaultPackage_ECUProjects',
                                                     'DefaultPackage_Blueprints',
                                                     'DefaultPackage_BlueprintMappingSets',
                                                     'DefaultPackage_SwAddrMethods']),
                            'GRAPHIC_OBJECT_FILTER': set(['ShowCalibrationObjects', 'UseGraphicEditors']) }

    def save(self, dest_dir, force = True):
        dest_file = os.path.join(dest_dir, 'ProfileSettings.xml')
        if force or not os.path.isfile(dest_file):
            lines = ['<?xml version="1.0" encoding="utf-8"?>']
            lines.append('<PROFILE>')
            lines.append(self.indent('<Version>1.0</Version>',1))
            lines.extend(self.indent(self.gen_sections_all(),1))
            lines.append('</PROFILE>')
            with open(dest_file, 'w') as fp:
                fp.write('\n'.join(lines))
                fp.write('\n')

    def gen_sections_all(self):
        lines = []
        for name in self.sectionKeys.keys():
            section_values = {}
            for key in self.sectionKeys[name]:
                value = getattr(self, key, None)
                if value is not None:
                    section_values[key]=value
            if len(section_values)>0: #prevent generating empty sections
                lines.extend(self.gen_section(name, section_values))
        return lines

    def gen_section(self, name, section_values):
        lines = []
        lines.append('<SECTION NAME="{}">'.format(name))
        for key in sorted(section_values.keys()):
            value = section_values[key]
            if isinstance(value, bool):
                value = str(int(value))
            lines.append(self.indent('<KEYVALPAIR KEY="{}" VAL="{}" />'.format(key, value),1))
        lines.append('</SECTION>')
        return lines

class Dcf(XMLWriterSimple):
    """
    Container for DaVinci Developer configuration file (DCF).
    DaVinci 3.x not supported
    """
    def __init__(self, ws, profile = None):
        """
        Constructor
        The profile argument should be is an instance of autosar.Workspace (only used when saving DCF files)
        """
        super().__init__()
        self.file_ref = []
        self.external_file_ref = []
        self.ws = ws

        self.component_name = None
        if profile is not None:
            if not isinstance(profile, DcfProfile):
                raise ValueError("profile must be an instance of DcfProfile")
            self.profile = profile
        else:
            self.profile = DcfProfile()

    def adjust_file_refs(self, basedir):
        for elem in self.file_ref+self.external_file_ref:
            self._adjust_elem(elem, basedir)

    def _adjust_elem(self, elem, basedir):
        basename = ntpath.basename(elem['path'])
        dirname = ntpath.normpath(ntpath.join(basedir,ntpath.dirname(elem['path'])))
        elem['path']=ntpath.join(dirname,basename)
        if os.path.sep == '/': #are we running in cygwin/Linux?
            elem['path'] = elem['path'].replace(r'\\','/')

    def loadReferences(self, ws = None, external = True):
        """
        Loads ARXML from referenced files into an AUTOSAR workspace.
        Returns the workspace object.
        Parameters:

        * ws: Workspace object where ARXML will be loaded
        * external: If True it will recursively load externally referenced DCF files (DCF inside DCF)

        """
        parser = DcfParser()
        if ws is None:
            ws = autosar.workspace()
        for xml_path in [x['path'] for x in self.file_ref ]:
            ws.loadXML(xml_path)
        if external:
            for external_dcf in self.external_file_ref:
                child_path = external_dcf['path']
                if os.path.exists(child_path):
                    root, ext = os.path.splitext(child_path)
                    if ext == '.dcf':
                        child_dcf = parser.parse(child_path)
                        child_dcf.loadReferences(ws)
                else:
                    print("No such file: "+child_path, file=sys.stderr)
        return ws

    def save(self, dest_dir, dcf_name, file_map = None, comp_dir = None, force = False, single_file = None):
        """
        Saves the DaVinci Configuration to file system.
        parameters:
            dest_dir (str): Destination directory (will be automatically created if not exists)
            dcf_name: Name of the generated DCF file
            file_map: A dictionary where key is the ARXML file and value is another dictionary with two keys ('root' and 'filters')
            comp_dir: Composition directory (reserved for future use)
            force: If true, Overwrites file(s) if it already exists.
            single_file: Name of single ARXML file. Disabled usage of file_map (Not yet implemented)

        Examples of file_map:
            file_map = {'AUTOSAR_Platform': {'root': 'DATATYPE', 'filters': ['/AUTOSAR_Platform']}}

        Valid root-names in DCF (you can use the same root-name multiple times in the same DCF):

        * 'COMPONENTTYPE'
        * 'CONSTANT'
        * 'DATATYPE'
        * 'METHOD'
        * 'PACKAGE'
        * 'PORTINTERFACE'

        """
        self.make_dirs(dest_dir)
        if single_file:
            file_map = self._create_single_file_map(single_file)
        else:
            if file_map is None:
                file_map = self.create_default_file_map()
        self.save_xml_from_file_map(dest_dir, file_map, force)
        self.profile.save(dest_dir, force)
        self.save_dcf(dest_dir, dcf_name, file_map, force = force)

    def save_xml_single(self, dest_dir, file_name, force):
        raise NotImplementedError("single XML")

    def save_xml_from_file_map(self, dest_dir, xml_file_map, force):
        for key in xml_file_map.keys():
            file_name = key
            if not file_name.lower().endswith('.arxml'):
                file_name+='.arxml'
            dest_file = os.path.join(dest_dir, file_name)
            if force or not os.path.isfile(dest_file):
                elem = xml_file_map[key]
                self.ws.saveXML(dest_file, filters=elem['filters'])

    def create_default_file_map(self):
        dataTypesPackage = self.ws.findRolePackage('DataType')
        constantsPackage = self.ws.findRolePackage('Constant')
        portInterfacePackage = self.ws.findRolePackage('PortInterface')
        modeDeclarationPackage = self.ws.findRolePackage('ModeDclrGroup')
        componentTypePackage = self.ws.findRolePackage('ComponentType')
        file_map = {}
        if dataTypesPackage is not None:
            file_map['DataTypes'] = {'root': 'DATATYPE', 'filters': [dataTypesPackage.ref]}
        if constantsPackage is not None:
            file_map['Constants'] = {'root': 'CONSTANT', 'filters': [constantsPackage.ref]}
        if (portInterfacePackage is not None) or (modeDeclarationPackage is not None):
            filters = []
            if portInterfacePackage is not None:
                filters.append(portInterfacePackage.ref)
            if modeDeclarationPackage is not None:
                filters.append(modeDeclarationPackage.ref)
            file_map['PortInterfaces'] = {'root': 'PORTINTERFACE', 'filters': filters}
        if componentTypePackage is not None:
            file_map['ComponentTypes'] = {'root': 'COMPONENTTYPE', 'filters': [componentTypePackage.ref]}
        return file_map

    def save_dcf(self, dest_dir, dcf_name, xml_file_map, schema = None, force = True):
        """Generates a new DCF file. Will not overwrite existing file unless force is true"""
        file_ext = '' if dcf_name.lower().endswith('.dcf') else '.dcf'
        dest_file = os.path.join(dest_dir, dcf_name+file_ext)
        if force or not os.path.isfile(dest_file):
            lines=['<?xml version="1.0" encoding="utf-8"?>']
            if schema is None:
                #TODO: Below line needs to be improved
                schema_string = "{0}{1}_DEV".format(str(self.ws.version).replace('.',''), self.ws.patch)
            else:
                schema_string = schema
            lines.append('<DCF ARSCHEMA="{}">'.format(schema_string))
            lines.append(self.indent('<Version>1.0</Version>',1))
            lines.append(self.indent('<NAME>{}</NAME>'.format(dcf_name),1))
            lines.append(self.indent('<PROFILESETTINGS>ProfileSettings.xml</PROFILESETTINGS>',1))
            for key in xml_file_map.keys():
                elem = xml_file_map[key]
                extension = '' if key.lower().endswith('.arxml') else '.arxml'
                file_name = key + extension
                lines.extend(self.indent(self._single_file_ref(file_name),1))
            lines.append('</DCF>')
            with open(dest_file, 'w') as fp:
                fp.write('\n'.join(lines))
                fp.write('\n')

    def _component_file_ref(self, component_name):
        lines = ['<FILEREF>']
        lines.append(self.indent('<ARXML ROOTITEM="COMPONENTTYPE" TYPE="">{}.arxml</ARXML>'.format(component_name),1))
        lines.append(self.indent('<DVG>{}.dvg</DVG>'.format(component_name),1))
        lines.append('</FILEREF>')
        return lines

    def _single_file_ref(self, file_name, directory=None):
        if directory is not None:
            file_path = os.path.join(directory, file_name)
        else:
            file_path = file_name
        lines = ['<FILEREF>']
        lines.append(self.indent('<ARXML>{}</ARXML>'.format(file_path),1))
        lines.append('</FILEREF>')
        return lines

class DcfParser:
    """
    Parser a DaVinci Configuration File (DCF).
    DaVinci 3.x not supported
    """

    def parse(self, filename):
        """
        Parses file references and external file references from a DCF file
        """
        basedir = ntpath.dirname(filename)
        xml_root = self._open_xml(filename)
        dcf = self._process_xml(xml_root)
        dcf.adjust_file_refs(basedir)
        return dcf

    def _open_xml(self, filename):
        xmltree = ElementTree.ElementTree()
        xmltree.parse(filename)
        xmlroot = xmltree.getroot();
        return xmlroot

    def _process_xml(self,xmlroot):
       dcf = Dcf()
       for elem in xmlroot.findall('./FILEREF'):
           node = elem.find('./ARXML')
           root_item = node.attrib['ROOTITEM'] if 'ROOTITEM' in node.attrib else None
           dcf.file_ref.append( {'rootItem': root_item, 'path': node.text} )
       for elem in xmlroot.findall('./EXTERNALFILEREF'):
           node = elem.find('./PATH')
           dcf.external_file_ref.append( {'path': node.text} )
       return dcf



import autosar.package
import autosar.parser.package_parser
import autosar.writer
from autosar.base import (parseXMLFile, getXMLNamespace, removeNamespace, parseAutosarVersionAndSchema, prepareFilter, parseVersionString)
import json
import os
import ntpath
import collections
import re
#default parsers
from autosar.parser.datatype_parser import (DataTypeParser, DataTypeSemanticsParser, DataTypeUnitsParser)
from autosar.parser.portinterface_parser import (PortInterfacePackageParser,SoftwareAddressMethodParser)
from autosar.parser.constant_parser import ConstantParser
from autosar.parser.behavior_parser import BehaviorParser
from autosar.parser.component_parser import ComponentTypeParser
from autosar.parser.system_parser import SystemParser
from autosar.parser.signal_parser import SignalParser
from autosar.parser.mode_parser import ModeDeclarationParser
from autosar.parser.swc_implementation_parser import SwcImplementationParser
#default writers
from autosar.writer.datatype_writer import XMLDataTypeWriter, CodeDataTypeWriter
from autosar.writer.constant_writer import XMLConstantWriter, CodeConstantWriter
from autosar.writer.portinterface_writer import XMLPortInterfaceWriter, CodePortInterfaceWriter
from autosar.writer.component_writer import XMLComponentTypeWriter, CodeComponentTypeWriter
from autosar.writer.behavior_writer import XMLBehaviorWriter, CodeBehaviorWriter
from autosar.writer.signal_writer import SignalWriter
from autosar.writer.mode_writer import XMLModeWriter

_validWSRoles = ['DataType', 'Constant', 'PortInterface', 'ComponentType', 'ModeDclrGroup', 'CompuMethod', 'Unit',
                 'BaseType', 'DataConstraint']

class PackageRoles(collections.UserDict):
    def __init__(self, data = None):
        if data is None:
            data = {'DataType': None,
             'Constant': None,
             'PortInterface': None,
             'ModeDclrGroup': None,
             'ComponentType': None,
             'CompuMethod': None,
             'Unit': None,
             'DataConstraint': None }
        super().__init__(data)

class WorkspaceProfile:
    """
    A Workspace profile allows users to customize default settings and behaviors
    """
    def __init__(self):
        self.compuMethodSuffix = ''
        self.dataConstraintSuffix = '_DataConstr'
        self.swCalibrationAccessDefault = 'NOT-ACCESSIBLE'
        self.modeSwitchEnhancedModeDefault = False
        self.modeSwitchSupportAsyncDefault = False
        self.modeSwitchAutoSetModeGroupRef = False
        self.swBaseTypeEncodingDefault = 'NONE'

class Workspace:
    """
    An autosar worspace
    """
    def __init__(self, version, patch, schema, release = None, attributes = None, useDefaultWriters=True):
        self.packages = []
        if isinstance(version, str):
            (major, minor, patch) = parseVersionString(version)
            self._version=float("%s.%s"%(major, minor))
            self.patch=patch
        elif isinstance(version, float):
            self._version=version
            self.patch=int(patch)
        self.release = None if release is None else int(release)
        self.schema=schema
        self.packageParser=None
        self.packageWriter=None
        self.xmlroot = None
        self.attributes = attributes
        self.useDefaultWriters = bool(useDefaultWriters)
        self.roles = PackageRoles()
        self.roleStack = collections.deque() #stack of PackageRoles
        self.map = {'packages': {}}
        self.profile = WorkspaceProfile()
        self.unhandledParser = set() # [PackageParser] unhandled:
        self.unhandledWriter =set() #[PackageWriter] Unhandled

    @property
    def version(self):
        return self._version

    @version.setter
    def version(self, version):
        if isinstance(version, str):
            (major, minor, patch) = parseVersionString(version)
            self._version=float("%s.%s"%(major, minor))
            self.patch=patch
        elif isinstance(version, float):
            self._version=version

    @property
    def version_str(self):
        if self.patch is None:
            return str(self._version)
        else:
            return str(self._version)+'.'+str(self.patch)

    def __getitem__(self,key):
        if isinstance(key,str):
            return self.find(key)
        else:
            raise ValueError('expected string')

    def _adjustFileRef(self,fileRef,basedir):
        basename = ntpath.basename(fileRef['path'])
        dirname=ntpath.normpath(ntpath.join(basedir,ntpath.dirname(fileRef['path'])))
        retval=ntpath.join(dirname,basename)
        if os.path.sep == '/': #are we running in cygwin/Linux?
            retval = retval.replace(r'\\','/')
        return retval

    def getRole(self, role):
        return self.roles[role]

    def setRole(self, ref, role):
        if (role is not None) and (role not in _validWSRoles):
            raise ValueError('Invalid role name: '+role)
        if ref is None:
            self.roles[role]=None
        else:
            package = self.find(ref)
            if package is None:
                raise ValueError('Invalid reference: '+ref)
            if not isinstance(package, autosar.package.Package):
                raise ValueError('Invalid type "%s"for reference "%s", expected Package type'%(str(type(package)),ref))
            package.role=role
            self.roles[role]=package.ref

    def setRoles(self, *items):
        """
        Same as setRole but caller gives a list of tuples where the first item is the package reference, and second item is the role name
        """
        for item in items:
            self.setRole(item[0], item[1])

    def pushRoles(self):
        """
        Saves current package role settings in internal role stack
        """
        self.roleStack.append(PackageRoles(self.roles))

    def popRoles(self):
        """
        Restores last saved package role settings
        """
        roles = self.roleStack.pop()
        self.roles.update(roles)

    def openXML(self,filename):
        xmlroot = parseXMLFile(filename)
        namespace = getXMLNamespace(xmlroot)

        assert (namespace is not None)
        (major, minor, patch, release, schema) = parseAutosarVersionAndSchema(xmlroot)
        removeNamespace(xmlroot,namespace)
        self.version=float('%s.%s'%(major,minor))
        self.major = major
        self.minor = minor
        self.patch = patch
        self.release = release
        self.schema = schema
        self.xmlroot = xmlroot
        if self.version < 3.0:
            raise NotImplementedError("Version below 3.0 is not supported")
        if self.packageParser is None:
            self.packageParser = autosar.parser.package_parser.PackageParser(self.version)
        self._registerDefaultElementParsers(self.packageParser)

    def loadXML(self, filename, roles=None):
        global _validWSRoles
        self.openXML(filename)
        self.loadPackage('*')
        if roles is not None:
            if not isinstance(roles, collections.abc.Mapping):
                raise ValueError('roles parameter must be a dictionary or Mapping')
            for ref,role in roles.items():
                self.setRole(ref,role)

    def loadPackage(self, packagename, role=None):
        found=False
        result=[]
        if self.xmlroot is None:
            raise ValueError("xmlroot is None, did you call loadXML() or openXML()?")
        if self.version >= 3.0 and self.version < 4.0:
            if self.xmlroot.find('TOP-LEVEL-PACKAGES'):
                for xmlPackage in self.xmlroot.findall('./TOP-LEVEL-PACKAGES/AR-PACKAGE'):
                    if self._loadPackageInternal(result, xmlPackage, packagename, role):
                        found = True

        elif self.version>=4.0:
            if self.xmlroot.find('AR-PACKAGES'):
                for xmlPackage in self.xmlroot.findall('.AR-PACKAGES/AR-PACKAGE'):
                    if self._loadPackageInternal(result, xmlPackage, packagename, role):
                        found = True

        else:
            raise NotImplementedError('Version %s of ARXML not supported'%self.version)
        if found==False and packagename != '*':
            raise KeyError('package not found: '+packagename)

        if (self.unhandledParser):
            print("[PackageParser] unhandled: %s" % (", ".join(self.unhandledParser)))
        return result

    def _loadPackageInternal(self, result, xmlPackage, packagename, role):
        name = xmlPackage.find("./SHORT-NAME").text
        found = False
        if packagename=='*' or packagename==name:
            found=True
            package = self.find(name)
            if package is None:
                package = autosar.package.Package(name, parent=self)
                self.packages.append(package)
                result.append(package)
                self.map['packages'][name] = package
            self.packageParser.loadXML(package,xmlPackage)
            self.unhandledParser = self.unhandledParser.union(package.unhandledParser)
            if (packagename==name) and (role is not None):
                self.setRole(package.ref, role)
        return found

    def find(self, ref, role=None):
        global _validWSRoles
        if ref is None: return None
        if (role is not None) and ( ref[0] != '/'):
            if role not in _validWSRoles:
                raise ValueError("unknown role name: "+role)
            if self.roles[role] is not None:
                ref=self.roles[role]+'/'+ref #appends the role packet name in front of ref

        if ref[0]=='/': ref=ref[1:] #removes initial '/' if it exists
        ref = ref.partition('/')
        if ref[0] in self.map['packages']:
            pkg = self.map['packages'][ref[0]]
            if len(ref[2])>0:
                return pkg.find(ref[2])
            return pkg
        return None

    def findall(self,ref):
        """
        experimental find-method that has some rudimentary support for globs.
        """
        if ref is None: return None
        if ref[0]=='/': ref=ref[1:] #removes initial '/' if it exists
        ref = ref.partition('/')
        if ref[0]=='*' and len(ref[2])==0:
            result=list(self.packages)
        else:
            result=[]
            for pkg in self.packages:
                if pkg.name == ref[0] or ref[0]=='*':
                    if len(ref[2])>0:
                        result.extend(pkg.findall(ref[2]))
                    else:
                        result.append(pkg)
        return result

    def findRolePackage(self, roleName):
        """
        Returns package with role set to roleName or None
        """
        if roleName is None: return None
        for pkg in self.packages:
            if pkg.role == roleName:
                return pkg
            elif len(pkg.subPackages)>0:
                for childPkg in pkg.subPackages:
                    if childPkg.role == roleName:
                        return childPkg
        return None

    def createPackage(self,name,role=None):
        if name not in self.map['packages']:
            package = autosar.package.Package(name,self)
            self.packages.append(package)
            self.map['packages'][name] = package
            if role is not None:
                self.setRole(package.ref, role)
            return package
        else:
            return self.map['packages'][name]

    def dir(self,ref=None,_prefix='/'):
        if ref is None:
            return [x.name for x in self.packages]
        else:
            if ref[0]=='/':
                ref=ref[1:]
            ref = ref.partition('/')
            result=self.find(ref[0])
            if result is not None:
                return result.dir(ref[2] if len(ref[2])>0 else None,_prefix+ref[0]+'/')
            else:
                return None

    def findWS(self):
        return self

    def rootWS(self):
        return self

    def saveXML(self, filename, filters=None, ignore=None):
        if self.packageWriter is None:
            self.packageWriter = autosar.writer.package_writer.PackageWriter(self.version, self.patch)
            if self.useDefaultWriters:
                self._registerDefaultElementWriters(self.packageWriter)
        workspaceWriter=autosar.writer.WorkspaceWriter(self.version, self.patch, self.schema, self.packageWriter)
        with open(filename, 'w', encoding="utf-8") as fp:
            if isinstance(filters,str): filters=[filters]
            if isinstance(ignore,str): filters=[ignore]
            if filters is not None:
                filters = [prepareFilter(x) for x in filters]
            workspaceWriter.saveXML(self, fp, filters, ignore)

        if (self.unhandledWriter):
            print( "[PackageWriter] unhandled: %s" % (", ".join(  self.unhandledWriter  )) )

    def toXML(self, filters=None, ignore=None):
        if self.packageWriter is None:
            self.packageWriter = autosar.writer.package_writer.PackageWriter(self.version, self.patch)
            if self.useDefaultWriters:
                self._registerDefaultElementWriters(self.packageWriter)
        workspaceWriter=autosar.writer.WorkspaceWriter(self.version, self.patch, self.schema, self.packageWriter)
        if isinstance(filters,str): filters=[filters]
        if isinstance(ignore,str): filters=[ignore]
        if filters is not None:
            filters = [prepareFilter(x) for x in filters]
        return workspaceWriter.toXML(self, filters, ignore)

    def append(self,elem):
        if isinstance(elem,autosar.package.Package):
            self.packages.append(elem)
            elem.parent=self
            self.map['packages'][elem.name] = elem
        else:
            raise ValueError(type(elem))

### BEGIN DEPRECATED SECTION (2019-11-07)
    def toCode(self, filters=None, packages=None, header=None, version=None, patch=None):
        if version is None:
            version = self.version
        if patch is None:
            patch = self.patch
        writer=autosar.writer.WorkspaceWriter(version, patch, None, self.packageWriter)
        if isinstance(filters,str): filters=[filters]
        if isinstance(packages,str): packages=[packages]
        if packages is not None:
            if filters is None:
                filters = []
            for package in packages:
                if package[-1]=='/':
                    filters.append(package+'*')
                else:
                    filters.append(package+'/*')
        if filters is not None:
            filters = [prepareFilter(x) for x in filters]
        return writer.toCode(self, filters ,str(header), ws.noDefault)

    def saveCode(self, filename, filters=None, packages=None, ignore=None, head=None, tail=None, module=False, template=False, version=None, patch=None):
        """
        saves the workspace as python code so it can be recreated later
        """
        if version is None:
            version = self.version
        if patch is None:
            patch = self.patch
        if self.packageWriter is None:
            self.packageWriter = autosar.writer.package_writer.PackageWriter(version, patch)
            if self.useDefaultWriters:
                self._registerDefaultElementWriters(self.packageWriter)
        writer=autosar.writer.WorkspaceWriter(version, patch, None, self.packageWriter)
        if isinstance(packages,str): packages=[packages]
        if isinstance(filters,str): filters=[filters]
        if isinstance(ignore,str): ignore=[ignore]
        if packages is not None:
            if filters is None:
                filters = []
            for package in packages:
                if package[-1]=='/':
                    filters.append(package+'*')
                else:
                    filters.append(package+'/*')
        if filters is not None:
            filters = [prepareFilter(x) for x in filters]

        with open(filename,'w', encoding="utf-8") as fp:
            writer.saveCode(self, fp, filters, ignore, head, tail, module, template)
#### END DEPRECATED SECTION

    @property
    def ref(self):
        return ''

    def listPackages(self):
        """returns a list of strings containg the package names of the opened XML file"""
        packageList=[]
        if self.xmlroot is None:
            raise ValueError("xmlroot is None, did you call loadXML() or openXML()?")
        if self.version >= 3.0 and self.version < 4.0:
            if self.xmlroot.find('TOP-LEVEL-PACKAGES'):
                for xmlPackage in self.xmlroot.findall('./TOP-LEVEL-PACKAGES/AR-PACKAGE'):
                    packageList.append(xmlPackage.find("./SHORT-NAME").text)
        elif self.version>=4.0:
            if self.xmlroot.find('AR-PACKAGES'):
                for xmlPackage in self.xmlroot.findall('.AR-PACKAGES/AR-PACKAGE'):
                    packageList.append(xmlPackage.find("./SHORT-NAME").text)
        else:
            raise NotImplementedError('Version %s of ARXML not supported'%self.version)
        return packageList

    def delete(self, ref):
        if ref is None: return
        if ref[0]=='/': ref=ref[1:] #removes initial '/' if it exists
        ref = ref.partition('/')
        for i,pkg in enumerate(self.packages):
            if pkg.name == ref[0]:
                if len(ref[2])>0:
                    return pkg.delete(ref[2])
                else:
                    del self.packages[i]
                    del self.map['packages'][ref[0]]
                    break

    def createAdminData(self, data):
        return autosar.base.createAdminData(data)

    def apply(self, template, **kwargs):
        """
        Applies template to this workspace
        """
        if len(kwargs) == 0:
            template.apply(self)
        else:
            template.apply(self, **kwargs)
        template.usageCount+=1



    def registerElementParser(self, elementParser):
        """
        Registers a custom element parser object
        """
        if self.packageParser is None:
            self.packageParser = autosar.parser.package_parser.PackageParser(self.version)
            self._registerDefaultElementParsers(self.packageParser)
        self.packageParser.registerElementParser(elementParser)

    def registerElementWriter(self, elementWriter):
        """
        Registers a custom element parser object
        """
        if self.packageWriter is None:
            self.packageWriter = autosar.writer.package_writer.PackageWriter(self.version, self.patch)
            if self.useDefaultWriters:
                self._registerDefaultElementWriters(self.packageWriter)
        self.packageWriter.registerElementWriter(elementWriter)

    def _registerDefaultElementParsers(self, parser):
        parser.registerElementParser(DataTypeParser(self.version))
        parser.registerElementParser(DataTypeSemanticsParser(self.version))
        parser.registerElementParser(DataTypeUnitsParser(self.version))
        parser.registerElementParser(PortInterfacePackageParser(self.version))
        parser.registerElementParser(SoftwareAddressMethodParser(self.version))
        parser.registerElementParser(ModeDeclarationParser(self.version))
        parser.registerElementParser(ConstantParser(self.version))
        parser.registerElementParser(ComponentTypeParser(self.version))
        parser.registerElementParser(BehaviorParser(self.version))
        parser.registerElementParser(SystemParser(self.version))
        parser.registerElementParser(SignalParser(self.version))
        parser.registerElementParser(SwcImplementationParser(self.version))

    def _registerDefaultElementWriters(self, writer):
        writer.registerElementWriter(XMLDataTypeWriter(self.version, self.patch))
        writer.registerElementWriter(XMLConstantWriter(self.version, self.patch))
        writer.registerElementWriter(XMLPortInterfaceWriter(self.version, self.patch))
        writer.registerElementWriter(XMLComponentTypeWriter(self.version, self.patch))
        writer.registerElementWriter(XMLBehaviorWriter(self.version, self.patch))
        writer.registerElementWriter(CodeDataTypeWriter(self.version, self.patch))
        writer.registerElementWriter(CodeConstantWriter(self.version, self.patch))
        writer.registerElementWriter(CodePortInterfaceWriter(self.version, self.patch))
        writer.registerElementWriter(CodeComponentTypeWriter(self.version, self.patch))
        writer.registerElementWriter(CodeBehaviorWriter(self.version, self.patch))
        writer.registerElementWriter(SignalWriter(self.version, self.patch))
        writer.registerElementWriter(XMLModeWriter(self.version, self.patch))



import os
import autosar.rte.partition
import cfile as C
import io
import autosar.base
import autosar.bsw.com

innerIndentDefault=3 #default indendation (number of spaces)

def _genCommentHeader(comment):
   lines = []
   lines.append('/*********************************************************************************************************************')
   lines.append('* %s'%comment)
   lines.append('*********************************************************************************************************************/')
   return lines

def _genCommentHeader2(comment):
   """
   Same as _genCommentHeader but returns a C sequence instead of raw strings
   """
   code = C.sequence()
   code.append(C.line('/*********************************************************************************************************************'))
   code.append(C.line('* %s'%comment))
   code.append(C.line('*********************************************************************************************************************/'))
   return code


class TypeGenerator:

   def __init__(self, partition, useDefaultTypes=True):
      self.partition = partition
      self.defaultTypes = {}
      if useDefaultTypes:
         self._initDefaultType()


   def generate(self, dest_dir = '.', file_name='Rte_Type.h'):
      """
      Generates Rte_Type.h
      Note: The last argument has been deprecated and is no longer in use
      """
      if self.partition.isFinalized == False:
         self.partition.finalize()
      file_path = os.path.join(dest_dir, file_name)
      with io.open(file_path, 'w', newline='\n') as fp:
         hfile=C.hfile(file_name)
         hfile.code.extend([C.line(x) for x in _genCommentHeader('Includes')])
         hfile.code.append(C.include("Std_Types.h"))
         hfile.code.append(C.blank())
         (basicTypes,complexTypes,modeTypes) = self.partition.types.getTypes()
         hfile.code.extend([C.line(x) for x in _genCommentHeader('Data Type Definitions')])
         hfile.code.append(C.blank())
         ws = self.partition.ws
         unusedDefaultTypes = self._findUnusedDefaultTypes(ws, basicTypes)

         first=True
         for ref in sorted(basicTypes)+sorted(complexTypes):
            dataType = ws.find(ref)
            if dataType is not None:
               typedef = None
               if first:
                  first=False
               else:
                  hfile.code.append(C.blank())
               hfile.code.append('#define Rte_TypeDef_%s'%dataType.name)
               if isinstance(dataType,autosar.datatype.BooleanDataType):
                  typedef = C.typedef('boolean', dataType.name)
                  hfile.code.append(C.statement(typedef))
               elif isinstance(dataType,autosar.datatype.IntegerDataType):
                  valrange = dataType.maxVal-dataType.minVal
                  bitcount = valrange.bit_length()
                  typename = dataType.name
                  basetype = self._typename(bitcount,dataType.minVal)
                  typedef = C.typedef(basetype, typename)
                  hfile.code.append(C.statement(typedef))
                  isUnsigned = True if basetype in ('uint8','uint16','uint32') else False
                  if isUnsigned:
                     minval=str(dataType.minVal)+'u'
                     maxval=str(dataType.maxVal)+'u'
                  else:
                     minval=str(dataType.minVal)
                     maxval=str(dataType.maxVal)
                  hfile.code.append('#define %s_LowerLimit ((%s)%s)'%(typename,typename,minval))
                  hfile.code.append('#define %s_UpperLimit ((%s)%s)'%(typename,typename,maxval))
                  if dataType.compuMethodRef is not None:
                     compuMethod = ws.find(dataType.compuMethodRef)
                     if compuMethod is not None:
                        lines1=[]
                        lines2=[]
                        if compuMethod.intToPhys is not None:
                           for elem in compuMethod.intToPhys.elements:
                              if elem.textValue:
                                 if isUnsigned:
                                    value = str(elem.upperLimit)+'u'
                                 else:
                                    value = str(elem.upperLimit)
                                 lines1.append('#define RTE_CONST_%s (%s)'%(elem.textValue,value))
                                 lines2.append('#define %s ((%s)%s)'%(elem.textValue,typename,value))
                        if len(lines2)>0:
                           tmp=lines1+[C.blank()]+lines2
                        else:
                           tmp=lines1
                        for line in tmp:
                           hfile.code.append(line)
                     else:
                        raise ValueError(dataType.compuMethodRef)
               elif isinstance(dataType, autosar.datatype.RecordDataType):
                  body = C.block(innerIndent=innerIndentDefault)
                  for elem in dataType.elements:
                     childType = ws.find(elem.typeRef, role='DataType')
                     body.append(C.statement(C.variable(elem.name, childType.name)))
                  struct = C.struct(None,body, typedef=dataType.name)
                  hfile.code.append(C.statement(struct))
               elif isinstance(dataType, autosar.datatype.StringDataType):
                  hfile.code.append('typedef uint8 %s[%d];'%(dataType.name, dataType.length+1))
               elif isinstance(dataType, autosar.datatype.ArrayDataType):
                  childType = ws.find(dataType.typeRef, role='DataType')
                  if childType is None:
                     raise ValueError('invalid type reference: '+dataType.typeRef)
                  hfile.code.append('typedef %s %s[%d];'%(childType.name, dataType.name, dataType.length))
               elif isinstance(dataType, autosar.datatype.RealDataType):
                  if dataType.encoding == 'DOUBLE':
                     platform_typename = 'float64'
                  else:
                     platform_typename = 'float32'
                  hfile.code.append('typedef %s %s;'%(platform_typename, dataType.name))
               else:
                  raise NotImplementedError(type(dataType))
                  #sys.stderr.write('not implemented: %s\n'%str(type(dataType)))
            else:
               raise ValueError(ref)

         if len(modeTypes)>0:
            lines=_genCommentHeader('Mode Types')
            tmp=[]
            hfile.code.extend(lines)
            first=True
            for ref in modeTypes:
               if first:
                  first=False
               else:
                  tmp.append(C.blank())
               modeType = ws.find(ref)
               hfile.code.append(C.statement(C.typedef('uint8', 'Rte_ModeType_'+modeType.name)))

               for i,elem in enumerate(modeType.modeDeclarations):
                  # define RTE_MODE_EcuM_Mode_POST_RUN ((Rte_ModeType_EcuM_Mode)0)
                  tmp.append(C.define('RTE_MODE_%s_%s'%(modeType.name,elem.name),'((Rte_ModeType_EcuM_Mode)%d)'%i))

            hfile.code.append(C.blank())
            hfile.code.extend(tmp)
         if len(unusedDefaultTypes)>0:
            hfile.code.append(C.blank(2))
            hfile.code.append(C.line('#ifndef RTE_SUPPRESS_UNUSED_DATATYPES'))
            for name in sorted(unusedDefaultTypes):
               hfile.code.append(C.blank())
               hfile.code.extend(self.defaultTypes[name])
            hfile.code.append(C.blank())
            hfile.code.append(C.line('#endif'))
         fp.write('\n'.join(hfile.lines()))
         fp.write('\n')


   def _initDefaultType(self):
      self.defaultTypes['Boolean']=C.sequence().extend([C.statement(C.typedef('boolean', 'Boolean'))])
      self.defaultTypes['UInt8']=C.sequence().extend([C.statement(C.typedef('uint8', 'UInt8')), C.define('UInt8_LowerLimit', '((UInt8)0u)'), C.define('UInt8_UpperLimit', '((UInt8)255u)')])
      self.defaultTypes['UInt16']=C.sequence().extend([C.statement(C.typedef('uint16', 'UInt16')), C.define('UInt16_LowerLimit', '((UInt16)0u)'), C.define('UInt16_UpperLimit', '((UInt16)65535u)')])
      self.defaultTypes['UInt32']=C.sequence().extend([C.statement(C.typedef('uint32', 'UInt32')), C.define('UInt32_LowerLimit', '((UInt32)0u)'), C.define('UInt32_UpperLimit', '((UInt32)4294967295u)')])
      self.defaultTypes['SInt8']=C.sequence().extend([C.statement(C.typedef('sint8', 'SInt8')), C.define('SInt8_LowerLimit', '((SInt8)-128)'), C.define('SInt8_UpperLimit', '((SInt8)127)')])
      self.defaultTypes['SInt16']=C.sequence().extend([C.statement(C.typedef('sint16', 'SInt16')), C.define('SInt16_LowerLimit', '((SInt16)-32768)'), C.define('SInt16_UpperLimit', '((SInt16)32767)')])
      self.defaultTypes['SInt32']=C.sequence().extend([C.statement(C.typedef('sint32', 'SInt32')), C.define('SInt32_LowerLimit', '((SInt32)-2147483648)'), C.define('SInt32_UpperLimit', '((SInt32)2147483647)')])

   def _findUnusedDefaultTypes(self, ws, typerefs):
      defaultTypeNames = set(self.defaultTypes.keys())
      usedTypeNames = set()

      for ref in typerefs:
         dataType = ws.find(ref)
         if dataType is None:
            raise ValueError('invalid type reference: '+ref)
         usedTypeNames.add(dataType.name)
      return defaultTypeNames-usedTypeNames


   @staticmethod
   def _typename(bitcount,minval):
      if bitcount <=8:
         return 'uint8' if minval >= 0 else 'sint8'
      elif bitcount <=16:
         return 'uint16' if minval >= 0 else 'sint16'
      elif bitcount <=32:
         return 'uint32' if minval >= 0 else 'sint32'
      elif bitcount <=64:
         return 'uint64' if minval >= 0 else 'sint64'
      else:
         raise ValueError(bitcount)

class RteGenerator:
   """
   Generates Rte.c based on partition. The prefix argument can be used to change both
   the file name and prefix name used for public function names
   """
   def __init__(self, partition, prefix='Rte', include=None, mode_switch=True, os_enable=True):
      self.partition=partition
      self.includes = [] #array of tuples, first element is the name of include header, second element is True if this is a sysinclude
      self.prefix=prefix
      self.com_component = None
      self.header_file_name = None
      self.data_elements = []
      self.extra_static_vars={}
      self.extra_public_functions={}
      self.extra_rte_start=C.sequence()
      self.mode_switch_enable=mode_switch
      self.os_enable = os_enable
      #self.com_access = {'receive': {}, 'send': {}}
      if include is not None:
         for elem in include:
            if isinstance(elem, str) or isinstance(elem, tuple):
               self.includes.append(elem)
            else:
               raise ValueError("include items must be of type str or tuple(str,boolean)")
      for component in partition.components:
         if isinstance(component.inner, autosar.bsw.com.ComComponent):
            if self.com_component is None:
               self.com_component = component
            else:
               raise RuntimeError("More than one Com component allowed in a partition")

   def generate(self, dest_dir='.'):
      if self.os_enable:
         self.extra_static_vars.update(self.partition.static_vars)      
      self._generate_header(dest_dir, 'RteApi.h')
      self._generate_source(dest_dir, 'RteApi.c')
   
   def _generate_header(self, dest_dir='.', file_name=None):
      if file_name is None:
         file_name = 'RteApi.h'      
      self.includes.append((file_name, False))
      file_path = os.path.join(dest_dir,file_name)
      with io.open(file_path, 'w', newline='\n') as fp:
         header = C.hfile(file_path)
         self._write_header_includes(header.code)
         self._write_header_public_func(header.code)
         for line in header.lines():
            fp.write(line)
            fp.write('\n')
         fp.write('\n')
         
   
   def _generate_source(self, dest_dir='.', file_name=None):
      if file_name is None:
         file_name = 'RteApi.c'
      file_path = os.path.join(dest_dir,file_name)
      with io.open(file_path, 'w', newline='\n') as fp:
         self._write_includes(fp)
         self._write_constants_and_typedefs(fp)
         self._write_local_vars(fp)
         self._write_public_funcs(fp)

   def _write_includes(self, fp):
      lines = _genCommentHeader('Includes')
      fp.write('\n'.join(lines)+'\n')
      code = C.sequence()
      for include in self.includes:
         code.append(C.include(*include))
      if self.com_component is not None:
         code.append(C.include(self.com_component.name+'.h'))
      if self.os_enable:
         code.append(C.include('os.h'))
      fp.write('\n'.join(code.lines())+'\n\n')

   def _write_constants_and_typedefs(self, fp):
      fp.write('\n'.join(_genCommentHeader('Constants and Types'))+'\n\n')

   def _write_local_vars(self, fp):
      fp.write('\n'.join(_genCommentHeader('Local Variables'))+'\n')
      code = C.sequence()
      for data_element in sorted(self.partition.data_element_map.values(), key=lambda x: x.symbol):
         var = C.variable(data_element.symbol, data_element.dataType.name, True)
         code.append(C.statement(var))
      for key in sorted(self.extra_static_vars.keys()):
         code.append(C.statement(self.extra_static_vars[key]))
      fp.write('\n'.join(code.lines())+'\n\n')

   def _write_public_funcs(self, fp):      
      fp.write('\n'.join(_genCommentHeader('Public Functions'))+'\n')
      self._write_rte_start(fp)
      if len(self.partition.upperLayerAPI.read)>0:
        self._genRead(fp, sorted(self.partition.upperLayerAPI.final['read'], key=lambda x: x.shortname))
      if len(self.partition.upperLayerAPI.write)>0:
        self._genWrite(fp, sorted(self.partition.upperLayerAPI.final['write'], key=lambda x: x.shortname))
      if len(self.partition.upperLayerAPI.receive)>0:
        self._genReceive(fp, sorted(self.partition.upperLayerAPI.final['receive'], key=lambda x: x.shortname))
      if len(self.partition.upperLayerAPI.send)>0:
        self._genSend(fp, sorted(self.partition.upperLayerAPI.final['send'], key=lambda x: x.shortname))
      #if len(self.partition.upperLayerAPI.call)>0:
      #  self._genCall(fp, sorted(self.partition.upperLayerAPI.final['call'], key=lambda x: x.shortname))
      if len(self.partition.upperLayerAPI.get)>0:
        self._genGet(fp, sorted(self.partition.upperLayerAPI.final['get'], key=lambda x: x.shortname))
      if len(self.partition.upperLayerAPI.setReadData)>0:
        self._genFunctionBodies(fp, sorted(self.partition.upperLayerAPI.final['setReadData'], key=lambda x: x.shortname))
      if len(self.partition.upperLayerAPI.setReadResult)>0:
        self._genFunctionBodies(fp, sorted(self.partition.upperLayerAPI.setReadResult.values(), key=lambda x: x.shortname))
      if self.mode_switch_enable and len(self.partition.mode_switch_functions)>0:
         self._genFunctionBodies(fp, [self.partition.mode_switch_functions[key] for key in sorted(self.partition.mode_switch_functions.keys())])
      if len(self.extra_public_functions)>0:
         self._genFunctionBodies(fp, [self.extra_public_functions[key] for key in sorted(self.extra_public_functions.keys())])

   def _write_rte_start(self, fp):
      func = C.function(self.prefix+'_Start', 'void')
      body = C.block(innerIndent=innerIndentDefault)
      self._write_init_values(body)
      if len(self.extra_rte_start)>0:
         body.extend(self.extra_rte_start)
      fp.write(str(func)+'\n')
      fp.write('\n'.join(body.lines())+'\n\n')

   def _write_init_values(self, body):
      for data_element in sorted(self.partition.data_element_map.values(), key=lambda x: x.symbol):
         if data_element.initValue is not None:
            init_str = autosar.constant.initializer_string(data_element.initValue)
            body.code.append(C.statement('%s = %s'%(data_element.symbol, init_str)))



   def _genRead(self, fp, prototypes):
      """Generates all Rte_Read functions"""
      for port_func in prototypes:
         body = C.block(innerIndent=innerIndentDefault)
         if port_func.data_element.com_access['Receive'] is not None:
            com_func = port_func.data_element.com_access['Receive']
            body.code.append(C.statement('return '+str(C.fcall(com_func.name, args=[port_func.proto.args[0].name]))))
         else:
            body.code.append(C.statement('*%s = %s'%(port_func.proto.args[0].name, port_func.data_element.symbol)))
            if port_func.data_element.result_var is not None:
               body.code.append(C.statement('return %s'%port_func.data_element.result_var.name))
            else:
               body.code.append(C.statement('return RTE_E_OK'))

         fp.write(str(port_func.proto)+'\n')
         fp.write('\n'.join(body.lines())+'\n\n')

   def _genWrite(self, fp, prototypes):
      for port_func in prototypes:
         hasComSignal = False
         body = C.block(innerIndent=innerIndentDefault)
         if port_func.data_element.symbol is not None:
            body.code.append(C.statement('%s = %s'%(port_func.data_element.symbol, port_func.proto.args[0].name)))
         if port_func.data_element.com_access['Send'] is not None:
            com_func = port_func.data_element.com_access['Send']
            body.code.append(C.statement('return '+str(C.fcall(com_func.name, args=[port_func.proto.args[0].name]))))
         else:
            if port_func.data_element.result_var is not None:
               body.code.append(C.statement('return %s'%port_func.data_element.result_var.name))
            else:
               body.code.append(C.statement('return RTE_E_OK'))
         fp.write(str(port_func.proto)+'\n')
         fp.write('\n'.join(body.lines())+'\n\n')

   def _genReceive(self, fp, prototypes):
      for proto in prototypes:
         body = C.block(innerIndent=innerIndentDefault)
         body.code.append(C.statement('return RTE_E_OK'))
         fp.write(str(proto.func)+'\n')
         fp.write('\n'.join(body.lines())+'\n\n')

   def _genSend(self, fp, prototypes):
      for proto in prototypes:
         body = C.block(innerIndent=innerIndentDefault)
         body.code.append(C.statement('return RTE_E_OK'))
         fp.write(str(proto.func)+'\n')
         fp.write('\n'.join(body.lines())+'\n\n')

   def _genCall(self, fp, prototypes):
      for proto in prototypes:
         body = C.block(innerIndent=innerIndentDefault)
         body.code.append(C.statement('return RTE_E_OK'))
         fp.write(str(proto.func)+'\n')
         fp.write('\n'.join(body.lines())+'\n\n')

   def _genGet(self, fp, prototypes):
      for port_func in prototypes:
         body = C.block(innerIndent=innerIndentDefault)
         prefix = '&' if port_func.data_element.dataType.isComplexType else ''
         suffix = '[0]' if isinstance(port_func.data_element.dataType, autosar.datatype.ArrayDataType) else ''
         body.code.append(C.statement('return %s%s%s'%(prefix, port_func.data_element.symbol, suffix)))
         fp.write(str(port_func.proto)+'\n')
         fp.write('\n'.join(body.lines())+'\n\n')
   
   def _genFunctionBodies(self, fp, prototypes):
      for func in prototypes:
         fp.write(str(func.proto)+'\n')
         fp.write('\n'.join(func.body.lines())+'\n\n')
   
   def _write_header_includes(self, code):      
      code.extend(_genCommentHeader2("INCLUDES"))      
      code.append(C.include('Rte_Type.h'))
      code.append(C.include('Rte.h'))
   
   def _write_header_public_func(self, code):
      code.append('')
      code.extend(_genCommentHeader2("PUBLIC FUNCTION PROTOTYPES"))      
      code.append(C.statement(C.function('Rte_Start', 'void')))
      if self.mode_switch_enable and len(self.partition.mode_switch_functions)>0:
         for func in [self.partition.mode_switch_functions[key] for key in sorted(self.partition.mode_switch_functions.keys())]:
            code.append(C.statement(func.proto))

         

class ComponentHeaderGenerator():
   def __init__(self, partition):
      self.partition = partition
      self.useMockedAPI=False

   def generate(self, destdir, mocked=None):
      if mocked is not None:
         self.useMockedAPI=bool(mocked)
      for component in self.partition.components:
         if not isinstance(component.inner, autosar.bsw.com.ComComponent):
            with io.open(os.path.join(destdir, 'Rte_%s.h'%component.inner.name), 'w', newline='\n') as fp:
               self._genComponentHeader(fp, component)

   def _genComponentHeader(self, fp, component):
      ws = component.inner.rootWS()
      assert(ws is not None)
      hfile=C.hfile(None, guard='RTE_%s_H'%(component.inner.name.upper()))
      hfile.code.append(C.include('Rte.h'))
      hfile.code.append(C.include('Rte_Type.h'))
      hfile.code.append(C.blank())
      lines = self._genInitValues(ws, component.inner.requirePorts+component.inner.providePorts)
      if len(lines)>0:
         hfile.code.extend([C.line(x) for x in _genCommentHeader('Init Values')])
         hfile.code.extend(lines)

      
      #Write API      
      num_funcs = sum(1 for x in component.clientAPI.get_all())
      if num_funcs>0:
         hfile.code.append(C.blank())
         hfile.code.extend([C.line(x) for x in _genCommentHeader('API Prototypes')])
         for func in component.clientAPI.get_all():
            assert func.proto is not None
            hfile.code.append(C.statement(func.proto))
      if len(component.clientAPI.final['read'])>0:
         hfile.code.append(C.blank())
         hfile.code.extend([C.line(x) for x in _genCommentHeader('Rte_Read_<p>_<d>')])
         hfile.code.extend([C.define(func.shortname, func.proto.name) for func in component.clientAPI.final['read']])
      if len(component.clientAPI.final['write'])>0:
         hfile.code.append(C.blank())
         hfile.code.extend([C.line(x) for x in _genCommentHeader('Rte_Write_<p>_<d>')])
         hfile.code.extend([C.define(func.shortname, func.proto.name) for func in component.clientAPI.final['write']])
      if len(component.clientAPI.final['receive'])>0:
         hfile.code.append(C.blank())
         hfile.code.extend([C.line(x) for x in _genCommentHeader('Rte_Receive_<p>_<d>')])
         hfile.code.extend([C.define(func.shortname, func.proto.name) for func in component.clientAPI.final['receive']])
      if len(component.clientAPI.final['send'])>0:
         hfile.code.append(C.blank())
         hfile.code.extend([C.line(x) for x in _genCommentHeader('Rte_Send_<p>_<d>')])
         hfile.code.extend([C.define(func.shortname, func.proto.name) for func in component.clientAPI.final['send']])
      if len(component.clientAPI.final['mode'])>0:
         hfile.code.append(C.blank())
         hfile.code.extend([C.line(x) for x in _genCommentHeader('Rte_Mode_<p>_<d>')])
         hfile.code.extend([C.define(func.shortname, func.proto.name) for func in component.clientAPI.final['mode']])
      if len(component.clientAPI.final['mode'])>0:
         hfile.code.append(C.blank())
         hfile.code.extend([C.line(x) for x in _genCommentHeader('Rte_Mode_<mode>')])
         hfile.code.extend([C.define(func.shortname, func.proto.name) for func in component.clientAPI.final['mode']])
      if len(component.clientAPI.final['calprm'])>0:
         hfile.code.append(C.blank())
         hfile.code.extend([C.line(x) for x in _genCommentHeader('Rte_Calprm_<name>')])
         hfile.code.extend([C.define(func.shortname, func.proto.name) for func in component.clientAPI.final['calprm']])
      if len(component.clientAPI.final['call'])>0:
         hfile.code.append(C.blank())
         hfile.code.extend([C.line(x) for x in _genCommentHeader('Rte_Call_<p>_<o> ')])
         hfile.code.extend([C.define(func.shortname, func.proto.name) for func in component.clientAPI.final['call']])
      if len(component.runnables)>0:
         for runnable in sorted(component.runnables, key=lambda x: x.symbol):
            tmp = self._writeRunnableProto(runnable)
            hfile.code.extend(tmp)
      fp.write('\n'.join(hfile.lines()))
      fp.write('\n')

   def _genInitValues(self, ws, ports):
      ports = sorted(ports, key=lambda port: port.name)
      code = C.sequence()
      for port in ports:
         for comspec in port.comspec:
            if isinstance(comspec, autosar.port.DataElementComSpec):
               if comspec.initValueRef is not None:
                  initValue = ws.find(comspec.initValueRef)
                  if isinstance(initValue, autosar.constant.Constant):
                     #in case the ref is pointing to a Constant (the parent), grab the child instance using .value
                     initValue=initValue.value
                  if initValue is not None:
                     dataType = ws.find(initValue.typeRef)
                     if dataType is not None:
                        prefix = 'Rte_InitValue_%s_%s'%(port.name, comspec.name)
                        code.extend(self._getInitValue(ws, prefix, initValue, dataType))
      return code

   def _getInitValue(self, ws, def_name, value, dataType):
      """
      returns a list or sequence
      """
      code = C.sequence()
      if isinstance(value, autosar.constant.IntegerValue):
         if dataType.minVal>=0:
            suffix='u'
         else:
            suffix=''
         code.append(C.define(def_name,'((%s)%s%s)'%(dataType.name, value.value,suffix)))
      elif isinstance(value, autosar.constant.StringValue):
         code.append(C.define(def_name,'"%s"'%(value.value)))
      elif isinstance(value, autosar.constant.BooleanValue):
         if value.value:
            text='((boolean) TRUE)'
         else:
            text='((boolean) FALSE)'
         code.append(C.define(def_name,text))
      elif isinstance(value, autosar.constant.RecordValue):
         for element in value.elements:
            prefix = '%s_%s'%(def_name, element.name)
            dataType = ws.find(element.typeRef)
            if dataType is not None:
               code.extend(self._getInitValue(ws, prefix, element, dataType))
      elif isinstance(value, autosar.constant.ArrayValue):
         pass
      else:
         raise NotImplementedError(type(value))
      return code

   def _writeRunnableProto(self, runnable):
      lines = []
      lines.extend([C.line(x) for x in _genCommentHeader('Runnable %s'%runnable.name)])
      lines.append(C.statement(runnable.prototype))
      lines.append(C.blank())
      return lines


class MockRteGenerator(RteGenerator):
   def __init__(self, partition, api_prefix='Rte', file_prefix = 'MockRte', include=None, mode_switch=False, os_enable=False):
      super().__init__(partition, api_prefix, include, mode_switch)      
      self.includes.append((file_prefix+'.h', False))      
      self.api_prefix = api_prefix
      self.file_prefix = file_prefix
      self.os_enable = os_enable
      self.typedefs={}
      for port in partition.unconnectedPorts():
         if isinstance(port, autosar.rte.base.ProvidePort):
            self._create_port_getter_api(port)
         else:
            if len(port.data_elements)>0 or len(port.operations)>0:
               self._create_port_setter_api(port)
         self.partition.upperLayerAPI.finalize()

   def generate(self, dest_dir):
      self._generateHeader(dest_dir)
      super()._generate_source(dest_dir, self.file_prefix+'.c')      

   def _create_port_getter_api(self, port):
      component = port.parent
      for data_element in port.data_elements:
         if "%s/%s"%(port.name, data_element.name) in component.data_element_port_access:
            self._create_data_element_getter(component, port, data_element)

   def _create_data_element_getter(self, component, port, data_element):
      data_type = data_element.dataType
      func_name='%s_GetWriteData_%s_%s_%s'%(self.prefix, component.name, port.name, data_element.name)
      short_name='%s_GetWriteData_%s_%s'%(self.prefix, port.name, data_element.name)
      suffix = '*' if data_type.isComplexType else ''
      proto=C.function(func_name, data_type.name+suffix)
      rte_func = autosar.rte.base.DataElementFunction(proto, port, data_element)
      #self._createPortVariable(component, port, data_element)
      var_name = self._createDataElementVariable(component, port, data_element)
      self.partition.upperLayerAPI.get[short_name] = autosar.rte.base.GetPortFunction(short_name, proto, data_element)

   def _create_port_setter_api(self, port):
      component = port.parent
      for data_element in port.data_elements:
         if "%s/%s"%(port.name, data_element.name) in component.data_element_port_access:
            self._create_data_element_setter(component, port, data_element)
      for operation in port.operations:
         key = "%s/%s"%(port.name, operation.name)
         if key in component.operation_port_access:
            self._create_operation_setter(component, port, operation, component.operation_port_access[key])

   def _create_data_element_setter(self, component, port, data_element):
      var_name = self._createDataElementVariable(component, port, data_element)
      port_func = autosar.rte.base.SetReadDataFunction(self.prefix, component, port, data_element, var_name)
      self.partition.upperLayerAPI.setReadData[port_func.shortname] = port_func                  
      port_func = autosar.rte.base.SetReadResultFunction(self.prefix, component, port, data_element)
      self.partition.upperLayerAPI.setReadResult[port_func.shortname]=port_func
      self.extra_static_vars[port_func.static_var.name]=port_func.static_var
      self.extra_rte_start.append(C.statement('%s = RTE_E_OK'%(data_element.result_var.name)))
      

   def _create_operation_setter(self, component, port, operation, port_access):
      func_name='%s_SetCallHandler_%s_%s_%s'%(self.prefix, component.name, port.name, operation.name)
      short_name='%s_SetCallHandler_%s_%s'%(self.prefix, port.name, operation.name)

      type_name = '%s_%s_ServerCallHandler_t'%(port.name, operation.name)
      var_name = 'm_ServerCallHandler_%s_%s_%s'%(component.name, port.name, operation.name)      
      port_func = port.portAPI['Rte_Call_%s_%s'%(port.name, operation.name)]
      tmp_proto = C.fptr.from_func(port_func.proto, type_name)

      self.typedefs[type_name] = 'typedef %s'%str(tmp_proto)
      proto = C.function(func_name, 'void', params=[C.variable('handler_func', type_name, pointer=True)])
      func = autosar.rte.base.SetCallHandlerFunction(short_name, proto, operation, var_name)
      self.extra_public_functions[short_name]=func
      static_var = C.variable(var_name, type_name, static=True, pointer=True)
      self.extra_static_vars[var_name]=static_var
      self.extra_rte_start.append(C.statement('%s = (%s*) 0'%(var_name, type_name)))
      body = self._createMockServerCallFunction(port_func.proto, var_name)
      self.extra_public_functions[port_func.proto.name]=autosar.rte.base.ServerCallFunction(port_func.proto, body)

   def _createMockServerCallFunction(self, proto, var_name):
      body = C.block(innerIndent=innerIndentDefault)
      body.append(C.line('if (%s != 0)'%(var_name)))
      inner = C.block(innerIndent=innerIndentDefault)
      fcall = C.fcall(var_name)
      for param in proto.params:
         fcall.add_arg(param.name)
      if proto.typename != 'void':
         inner.append(C.statement('return %s'%str(fcall)))
      else:
         inner.append(C.statement(fcall))
      body.append(inner)
      if proto.typename != 'void':
         body.append(C.statement('return RTE_E_OK'))
      return body


   def _createDataElementVariable(self, component, port, data_element):
      data_element_map = self.partition.data_element_map
      variable_name = '_'.join([component.name, port.name, data_element.name])
      if variable_name not in data_element_map:
         data_element.symbol = variable_name
         data_element_map[variable_name] = data_element
      return variable_name

   def _generateHeader(self, dest_dir):
      filepath = os.path.join(dest_dir,self.file_prefix+'.h')
      with io.open(filepath, 'w', newline='\n') as fp:
         for line in self._createHeaderLines(filepath):
            fp.write(line)
            fp.write('\n')
         fp.write('\n')


   def _createHeaderLines(self, filepath):
      hfile = C.hfile(filepath)
      code = hfile.code
      code.extend([C.line(x) for x in _genCommentHeader('Includes')])
      code.append(C.include("Std_Types.h"))
      code.append(C.include("Rte_Type.h"))
      code.append(C.include("Rte.h"))
      code.append(C.blank())
      code.extend(_genCommentHeader('Constants and Types'))
      for key in sorted(self.typedefs.keys()):
         code.append(C.statement(self.typedefs[key]))
      code.append(C.blank())
      code.extend([C.line(x) for x in _genCommentHeader('Public Function Declarations')])
      code.append(C.blank())
      
      code.append(C.statement(C.function('%s_Start'%self.api_prefix, 'void')))
      for func in sorted(self.partition.upperLayerAPI.final['get'], key=lambda x: x.shortname):
         assert func.proto is not None
         hfile.code.append(C.statement(func.proto))
      for func in sorted(self.partition.upperLayerAPI.final['setReadData'], key=lambda x: x.shortname):
         assert func.proto is not None
         hfile.code.append(C.statement(func.proto))
      for func in sorted(self.partition.upperLayerAPI.final['setReadResult'], key=lambda x: x.shortname):
         assert func.proto is not None
         hfile.code.append(C.statement(func.proto))
      for func in sorted(self.extra_public_functions.values(), key=lambda x: x.shortname):
         assert func.proto is not None
         hfile.code.append(C.statement(func.proto))
      hfile.code.append(C.blank())
      return hfile.lines()


class RteTaskGenerator:
   """
   RteTask C code generator
   """
   def __init__(self, partition, os_cfg, prefix='RteTask', include=None):
      self.partition = partition
      self.prefix = prefix
      self.os_cfg = os_cfg
      self.includes = [
                     #array of tuples, first element is the name of include header, second element is True if this is a sysinclude
                     ('stdio.h', True),
                     ('Rte.h', False),
                     ('Rte_Type.h', False),
                     #('%s.h'%self.prefix, False),
                     ('os.h', False),
         ]
      for component in self.partition.components:
         if not isinstance(component.inner, autosar.bsw.com.ComComponent):
            self.includes.append(('Rte_%s.h'%component.name, False))
      if include is not None:
         for elem in List(include):
            if isinstance(elem, str):
               self.includes.append((elem, False))
            elif isinstance(elem, tuple):
               self.includes.append(elem)
            else:
               raise ValueError("elem: expected string or tuple, got "+str(type(elem)))      

   def generate(self, dest_dir='.'):
      #self._generate_header(dest_dir)
      self._generate_source(dest_dir)
   
   def _generate_source(self, dest_dir):
      file_name = self.prefix+'.c'
      file_path = os.path.join(dest_dir,file_name)

      with open(file_path, 'w', newline='\n') as fp:         
         s1 = self._write_source_includes()
         s2 = self._write_source_constants_and_typedefs()
         s3 = self._write_source_local_funcs()
         s4 = self._write_source_global_funcs()
         for seq in [s1, s2, s3, s4]:
            fp.write('\n'.join(seq.lines())+'\n')
         
         

   def _write_source_includes(self):
      code = C.sequence()
      code.extend(_genCommentHeader2('INCLUDES'))
      code.append(C.blank())
      for include in self.includes:
         code.append(C.include(*include))
      code.append(C.blank())
      return code

   def _write_source_constants_and_typedefs(self):
      code = C.sequence()
      code.extend(_genCommentHeader2('CONSTANTS AND DATA TYPES'))
      code.append(C.blank())
      return code

   def _write_source_local_funcs(self):
      code = C.sequence()
      code.extend(_genCommentHeader2('LOCAL FUNCTION PROTOTYPES'))
      code.append(C.blank())
      return code

   def _write_source_global_funcs(self):
      code = C.sequence()
      code.extend(_genCommentHeader2('GLOBAL FUNCTIONS'))
      code.append(C.blank())      
      for task in sorted(self.os_cfg.tasks, key=lambda x: x.name):         
         code.append(C.line('OS_TASK_HANDLER({0.name}, arg)'.format(task)))
         code.append(self._generate_task_body(task))
         code.append(C.blank(2))
      return code

   def _generate_task_body(self, task):
      code = C.block(innerIndent=innerIndentDefault)
      isRunning=C.variable('isRunning', 'boolean')
      code.append(C.statement('{0} = TRUE'.format(str(isRunning))))
      code.append(C.statement('os_task_t *self = (os_task_t*)arg'))
      code.append('')
      code.append(C.line('if (self == 0)'))
      body = C.block(innerIndent=innerIndentDefault)
      body.append(C.statement('THREAD_RETURN(1)'))
      code.append(body)
      code.append('')
      code.append(C.line('while (isRunning == TRUE)'))
            
      while_block = C.block(innerIndent=innerIndentDefault)
      
      
      while_block.append(C.statement('uint32 eventMask'))
      while_block.append(C.statement('int8_t result = os_task_waitEvent(self, &eventMask)'))
      while_block.append(C.line('if (result == 0)'))
      if_block = C.block(innerIndent=innerIndentDefault)
      if_block.extend(self._generate_event_mask_triggers(task))
      while_block.append(if_block)
      while_block.append(C.line('else if(result > 0)'))      
      if_block = C.block(innerIndent=innerIndentDefault)
      if_block.append(C.statement('printf("%s_QuitEvent\\n")'%task.name))
      if_block.append(C.statement('isRunning = false'))
      while_block.append(if_block)
      while_block.append(C.line('else'))
      if_block = C.block(innerIndent=innerIndentDefault)
      if_block.append(C.statement(r'fprintf(stderr, "os_task_waitEvent failed\n")'))
      while_block.append(if_block)

      code.append(while_block)
      code.append('')
      code.append(C.statement('THREAD_RETURN(0)'))
      return code

   def _generate_event_mask_triggers(self, task):
      code = C.sequence()
      for runnable in task.runnables:
         if runnable.processed:
            continue
         matching_runnables = self._find_compatible_runnables(task, runnable)
         self._generate_runnable_calls(code, matching_runnables)
         for matching in matching_runnables:
            matching.processed=True
      return code

   def _find_compatible_runnables(self, task, current):
      result = [current]
      for other in task.runnables:
         if (other is not current) and (not other.processed):
            if len(current.event_triggers) == len(other.event_triggers):
               is_compatible = True
               for current_event in current.event_triggers:
                  found = False
                  for other_event in other.event_triggers:
                     if current_event.symbol == other_event.symbol:
                        found = True
                        break
                  if not found:
                     is_compatible = False
                     break
               if is_compatible:
                  result.append(other)
      return result
   
   def _generate_runnable_calls(self, code, matching_runnables):
      events = matching_runnables[0].event_triggers
      if len(events) == 1:
         event = events[0]
         if not isinstance(event, autosar.rte.base.OperationInvokedEvent):
            code.append(C.line('if (eventMask & %s)'%event.symbol))
            block = C.block(innerIndent = innerIndentDefault)
            for runnable in matching_runnables:
               block.append(C.statement(C.fcall(runnable.symbol)))
            code.append(block)
      elif len(events) > 1:      
         raise NotImplementedError('multiple events')
      
   def _generate_header(self, dest_dir):
      file_name = self.prefix+'.h'
      file_path = os.path.join(dest_dir,file_name)
      with io.open(file_path, 'w', newline='\n') as fp:
         print("#ifndef RTE_TASK_H", file=fp)
         print("#define RTE_TASK_H", file=fp)
         self._write_header_includes(fp)
         self._write_header_constants_and_typedefs(fp)
         self._write_header_global_var(fp)
         self._write_header_global_proto(fp)
         print("#endif //RTE_TASK_H", file=fp)
   
   def _write_header_includes(self, fp):
      lines = _genCommentHeader('INCLUDES')
      lines.append('#ifdef _MSC_VER')
      lines.append('#include <Windows.h>')
      lines.append('#else')
      lines.append('#include <pthread.h>')
      lines.append('#endif //_MSC_VER')
      lines.append('#include "osmacro.h"')
      fp.write('\n'.join(lines)+'\n\n')

   def _write_header_constants_and_typedefs(self, fp):
      lines = _genCommentHeader('CONSTANTS AND DATA TYPES')
      fp.write('\n'.join(lines)+'\n\n')

   def _write_header_global_var(self, fp):
      lines = _genCommentHeader('GLOBAL VARIABLES')
      fp.write('\n'.join(lines)+'\n\n')

   def _write_header_global_proto(self, fp):
      lines = _genCommentHeader('GLOBAL FUNCTION PROTOTYPES')
      for task in self.os_cfg.tasks:
         lines.append('THREAD_PROTO(%s, arg);'%task.name)
      fp.write('\n'.join(lines)+'\n\n')
